# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Van der Waals radii

Available sets:

    - csd:  extrapolated from the Cambridge Structural Database by Santiago
    Alvarez, as seen in S. Alvarez, "A cartography of the van der Waals
    territories", Dalton Trans. 42, 8617 (2013)
    - jmol: extracted from the source code of JMol
    - ase:  default set for the Atomic Simulation Environment

"""

import json
import pkgutil

import numpy as np
from ase.data import atomic_numbers
from ase.data.vdw import vdw_radii as _vdw_radii_ase


def _load_vdw(name):

    _vdw_data = pkgutil.get_data("soprano", f"data/vdw_{name}.json").decode(
        "utf-8"
    )
    _vdw_radii = np.array(json.loads(_vdw_data))

    return _vdw_radii


vdw_radii = {"ase": _vdw_radii_ase, "jmol": _load_vdw("jmol"), "csd": _load_vdw("csd")}


def vdw_radius(el, vdwset="csd"):
    """Return Van der Waals radius for a certain element

    Return a Van der Waals radius for a given element, as taken from one
    of three available databases:

    - csd:  extrapolated from the Cambridge Structural Database by Santiago
    Alvarez, as seen in S. Alvarez, "A cartography of the van der Waals
    territories", Dalton Trans. 42, 8617 (2013) [default]
    - jmol: extracted from the source code of JMol
    - ase:  default set for the Atomic Simulation Environment

    | Args:
    |   el (str):       element symbol
    |   vdwset (str):   VdW set to use. Default is 'csd'.

    | Returns:
    |   vdw_radius (float): the Van der Waals radius
    """

    try:
        Z = atomic_numbers[el]
    except KeyError:
        raise ValueError("Invalid element symbol")

    return vdw_radii[vdwset][Z]


def build_custom_vdw(vdw_set, vdw_scale=1.0, default_vdw=2.0, vdw_custom={}):
    """Build a custom VdW set"""

    vdw_r = np.array(vdw_radii[vdw_set]) * vdw_scale
    vdw_r = np.where(np.isnan(vdw_r), default_vdw, vdw_r)
    for el, r in vdw_custom.items():
        vdw_r[atomic_numbers[el]] = r

    return vdw_r



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""CLI wrapper around the ASE GUI to view structures.

This adds a few options to the ASE GUI, and allows for the use of
Soprano's selection syntax to select atoms to be highlighted/hidden.

It will also check if the material seems to be an organic crystal and, if so, 
'reload' it with the molecule components joined up correctly.

TODO: re-order the operations so that the reload is done first, then the selection.
 - allow tagging by symmetry/labels even when not reducing the structure.

"""

__author__ = "J. Kane Shenton"
__maintainer__ = "J. Kane Shenton"
__email__ = "kane.shenton@stfc.ac.uk"
__date__ = "Nov 15, 2023"


import logging

import click
import click_log
from ase import Atoms
from ase.io import read

from soprano.scripts.cli_utils import (
    VIEW_OPTIONS,
    add_options,
    reload_as_molecular_crystal,
    viewimages,
)
from soprano.scripts.nmr import nmr_extract_atoms

# logging
logging.captureWarnings(True)
logger = logging.getLogger("cli")
click_log.basic_config(logger)

@click.command()
# one of more files
@click.argument("files", nargs=-1, type=click.Path(exists=True), required=True)
# verbosity flag e.g. -v -vv -vvv
@click.option(
    "--verbosity",
    "-v",
    count=True,
    help="Increase verbosity. "
    "Use -v for info, -vv for debug. Defaults to only showing warnings and errors.",
)
@add_options(VIEW_OPTIONS)
def view(
    files,
    subset,
    reduce,
    average_group,
    symprec,
    verbosity,
):
    """
    Visualise the structure(s) in the given file(s) using the ASE GUI.

    The user can select atoms to be tagged/hidden using the Soprano
    selection syntax.
    """



    # Set up logging
    if verbosity == 0:
        logging.basicConfig(level=logging.WARNING)
    elif verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.DEBUG)

    images = []
    for f in files:
        # get the atoms objects
        atoms_list = read(f, ":")
        if isinstance(atoms_list, Atoms):
            atoms_list = [atoms_list]

        for atoms in atoms_list:
            # check if the material is an organic crystal
            # if so, join up the molecule components
            atoms = reload_as_molecular_crystal(atoms, force=False)
            # get the subset of atoms to be tagged/hidden
            atoms = nmr_extract_atoms(
                atoms,
                subset=subset,
                reduce=reduce,
                average_group=average_group,
                symprec=symprec,
                logger=logger,
            )
            # add the atoms to the list
            images.append(atoms)

    # view the images
    viewimages(images)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Module containing AtomsProperties that relate to NMR properties of a system.
Some of these are valid only for Atoms objects loaded from a .magres file.
"""


from soprano.properties.nmr.dipolar import (
    DipolarCoupling,
    DipolarDiagonal,
    DipolarEuler,
    DipolarRSS,
    DipolarTensor,
)
from soprano.properties.nmr.efg import (
    EFGNQR,
    EFGAnisotropy,
    EFGAsymmetry,
    EFGDiagonal,
    EFGEuler,
    EFGQuadrupolarConstant,
    EFGQuadrupolarProduct,
    EFGQuaternion,
    EFGReducedAnisotropy,
    EFGSkew,
    EFGSpan,
    EFGTensor,
    EFGVzz,
)
from soprano.properties.nmr.isc import (
    JCDiagonal,
    JCIsotropy,
)
from soprano.properties.nmr.ms import (
    MSAnisotropy,
    MSAsymmetry,
    MSDiagonal,
    MSEuler,
    MSIsotropy,
    MSQuaternion,
    MSReducedAnisotropy,
    MSSkew,
    MSSpan,
    MSTensor,
)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
NMR Data

Data on NMR relevant properties of elements and isotopes - spin, gyromagnetic
ratio and quadrupole moment.
"""

import json
import pkgutil
import re

import numpy as np
import scipy.constants as cnst

# EFG conversion constant.
# Units chosen so that EFG_TO_CHI*Quadrupolar moment*Vzz = Hz
EFG_TO_CHI = (
    cnst.physical_constants["atomic unit of electric field " "gradient"][0]
    * cnst.e
    * 1e-31
    / cnst.h
)

try:
    _nmr_data = pkgutil.get_data("soprano", "data/nmrdata.json").decode("utf-8")
    _nmr_data = json.loads(_nmr_data)
except OSError:
    _nmr_data = None


def _get_nmr_data():

    if _nmr_data is not None:
        return _nmr_data
    else:
        raise RuntimeError(
            "NMR data not available. Something may be "
            "wrong with this installation of Soprano"
        )

def _get_isotope_list(elems, isotopes=None, isotope_list=None, use_q_isotopes=False):
    '''
    elems can be a single element string or a list of elements
    returns the isotope number for each element elems
    '''
    if isotopes is None:
        isotopes = {}
    if isinstance(elems, str):
        elems = [elems]  # It's a single element

    isotopelist = np.zeros(len(elems), dtype=int)
    nmr_data = _get_nmr_data()

    for i, e in enumerate(elems):

        if e not in nmr_data:
            # Non-existing element
            raise RuntimeError(f"No NMR data on element {e}")

        iso = nmr_data[e]["iso"]
        if use_q_isotopes and nmr_data[e]["Q_iso"] is not None:
            iso = nmr_data[e]["Q_iso"]
        if e in isotopes:
            iso = isotopes[e]
        if isotope_list is not None and isotope_list[i] is not None:
            iso = isotope_list[i]
        isotopelist[i] = iso
    return isotopelist

def _get_isotope_data(elems, key, isotopes={}, isotope_list=None, use_q_isotopes=False):
    isotopelist = _get_isotope_list(elems, isotopes=isotopes, isotope_list=isotope_list, use_q_isotopes=use_q_isotopes)

    data = np.zeros(len(elems))
    nmr_data = _get_nmr_data()

    for i, iso in enumerate(isotopelist):
        el = elems[i]
        try:
            data[i] = nmr_data[el][str(iso)][key]
        except KeyError:
            raise RuntimeError(
                f"Data {key} does not exist for isotope {iso} of "
                f"element {el}.\n"
                "Edit the file soprano/data/nmrdata.json to add custom isotopes."
            )

    return data


def _el_iso(sym):
    """Utility function: split isotope and element in conventional
    representation.
    """

    nmr_data = _get_nmr_data()

    match = re.findall("([0-9]*)([A-Za-z]+)", sym)
    if len(match) != 1:
        raise ValueError("Invalid isotope symbol")
    elif match[0][1] not in nmr_data:
        raise ValueError("Invalid element symbol")

    el = match[0][1]
    # What about the isotope?
    iso = str(nmr_data[el]["iso"]) if match[0][0] == "" else match[0][0]

    if iso not in nmr_data[el]:
        raise ValueError(f"No data on isotope {iso} for element {el}")

    return el, iso


def nmr_gamma(el, iso=None):
    """Gyromagnetic ratio for an element

    Return the gyromagnetic ratio for the given element and isotope, in
    rad/(s*T)

    | Args:
    |   el (str):   element symbol
    |   iso (int):  isotope. Default is the most abundant one.

    | Returns:
    |   gamma (float):  gyromagnetic ratio in rad/(s*T)
    """

    isotopes = {}
    if iso is not None:
        isotopes[el] = iso

    return _get_isotope_data([el], "gamma", isotopes=isotopes)[0]


def nmr_spin(el, iso=None):
    """Nuclear spin for an element

    Return the nuclear spin for the given element and isotope, in
    Bohr magnetons

    | Args:
    |   el (str):   element symbol
    |   iso (int):  isotope. Default is the most abundant one.

    | Returns:
    |   I (float):  nuclear spin in Bohr magnetons
    """

    isotopes = {}
    if iso is not None:
        isotopes[el] = iso

    return _get_isotope_data([el], "I", isotopes=isotopes)[0]


def nmr_quadrupole(el, iso=None):
    """Quadrupole moment for an element

    Return the quadrupole moment for the given element and isotope, in
    millibarn

    | Args:
    |   el (str):   element symbol
    |   iso (int):  isotope. Default is the most abundant one.

    | Returns:
    |   Q (float):  quadrupole moment in  millibarn
    """

    isotopes = {}
    if iso is not None:
        isotopes[el] = iso

    return _get_isotope_data([el], "Q", isotopes=isotopes)[0]



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Contains the NMRTensor class, simplifying the process of diagonalisation of an
NMR tensor as well as its representation in multiple conventions
"""

import warnings
from collections import defaultdict
from typing import Any, NamedTuple, Optional, TypeVar, Union

import numpy as np
from ase.quaternions import Quaternion
from numpy.lib.mixins import NDArrayOperatorsMixin
from scipy.spatial.transform import Rotation

from soprano.data.nmr import EFG_TO_CHI, _get_isotope_data, nmr_gamma, nmr_quadrupole, nmr_spin
from soprano.nmr.utils import (
    _anisotropy,
    _asymmetry,
    _dip_constant,
    _equivalent_euler,
    _equivalent_relative_euler,
    _evals_sort,
    _evecs_2_quat,
    _frange,
    _haeb_sort,
    _handle_euler_edge_cases,
    _matrix_to_euler,
    _skew,
    _span,
    _split_species,
    _test_euler_rotation,
    _tryallanglestest,
)

DEGENERACY_TOLERANCE = 1e-6

T = TypeVar("T", bound="NMRTensor")

class NMRTensor(NDArrayOperatorsMixin):
    """NMRTensor

    Class containing an NMR tensor, useful to access all its most important
    properties and representations.
    """

    ORDER_INCREASING = "i"
    ORDER_DECREASING = "d"
    ORDER_HAEBERLEN = "h"
    ORDER_NQR = "n"

    def __init__(self,
                 data: Union[np.ndarray, tuple[np.ndarray, np.ndarray]],
                 order: str=ORDER_INCREASING):
        """
        Initialise the NMRTensor

        Create an NMRTensor object from a 3x3 matrix.

        Arguments:
            data (np.ndarray or tuple):  3x3 matrix containing the tensor, or
                                         pair [evals, evecs] for the symmetric
                                         part alone.
            order (str):        Order to use for eigenvalues/eigenvectors. Can
                                be 'i' (ORDER_INCREASING), 'd'
                                (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                                'n' (ORDER_NQR). Default is 'i'.
        """
        self._order = None
        self._data = None
        self._symm = None
        self._evals = None
        self._evecs = None
        self._process_data(data)
        # The following will also sort the eigenvalues and eigenvectors
        self.order = order

        # Initialize other attributes
        self._anisotropy = None
        self._redaniso = None
        self._asymmetry = None
        self._span = None
        self._skew = None
        self._trace = None
        self._degeneracy = None
        # Spherical tensor components
        self._sph = None
        self._quat = None

        self._incr_evals = self._evals
        self._haeb_evals = _haeb_sort([self._evals])[0]

    def _process_data(self, data):
        if len(data) == 3:
            self._data = np.array(data)
            if self._data.shape != (3, 3):
                raise ValueError("Invalid matrix data passed to NMRTensor")
            self._symm = (self._data + self._data.T) / 2.0
            evals, evecs = np.linalg.eigh(self._symm)
        elif len(data) == 2:
            evals, evecs = data
            evecs = np.array(evecs)
            if len(evals) != 3 or evecs.shape != (3, 3):
                raise ValueError("Invalid eigenvalues/vectors passed to NMRTensor")
            self._symm = np.linalg.multi_dot([evecs, np.diag(evals), evecs.T])
            self._data = self._symm
        else:
            raise ValueError("Data must be a 3x3 matrix or a pair of [evals, evecs]")
        self._evals = evals
        self._evecs = evecs

    def _order_tensor(self, order):
        # Sort eigenvalues and eigenvectors as specified
        if self._order is None or self._order != order:
            self._evals, sort_i = _evals_sort([self._evals], order, True)
            self._evals = self._evals[0]
            self._evecs = self._evecs[:, sort_i[0]]
        # Last eigenvector must be the cross product of the first two
        self._evecs[:, 2] = np.cross(self._evecs[:, 0], self._evecs[:, 1])

        # For any property that depends on the eigenvalue order, reset it
        self._anisotropy = None
        self._redaniso = None
        self._asymmetry = None
        self._quat = None

    @property
    def order(self):
        return self._order
    # method to update the order of the tensor
    @order.setter
    def order(self, value):
        self._order_tensor(value)
        self._order = value

    @property
    def data(self):
        return self._data.copy()

    @property
    def eigenvalues(self):
        return self._evals.copy()

    @property
    def PAS(self):
        '''
        Returns the principal axis system (PAS) of the tensor.
        '''
        return np.diag(self._evals)

    @property
    def degeneracy(self):
        '''
        Returns the degeneracy of the tensor.
        For example, a tensor with eigenvalues [1, 1, 1] has a degeneracy of 3.
        A tensor with eigenvalues [1, 1, 2] has a degeneracy of 2.
        A tensor with eigenvalues [1, 2, 3] has a degeneracy of 1.
        '''
        if self._degeneracy is None:
            self._degeneracy = np.sum(np.abs(self._evals - self._evals[0]) < DEGENERACY_TOLERANCE)
        return self._degeneracy

    @property
    def eigenvectors(self):
        return self._evecs.copy()

    @property
    def trace(self):
        if self._trace is None:
            self._trace = np.trace(self._data)
        return self._trace

    @property
    def isotropy(self):
        return self.trace / 3.0

    @property
    def anisotropy(self):
        if self._anisotropy is None:
            self._anisotropy = _anisotropy(self._haeb_evals[None, :])[0]
        return self._anisotropy

    @property
    def reduced_anisotropy(self):
        if self._redaniso is None:
            self._redaniso = _anisotropy(self._haeb_evals[None, :], True)[0]
        return self._redaniso

    @property
    def asymmetry(self):
        if self._asymmetry is None:
            self._asymmetry = _asymmetry(self._haeb_evals[None, :])[0]
        return self._asymmetry

    @property
    def span(self):
        if self._span is None:
            self._span = _span(self._incr_evals[None, :])[0]
        return self._span

    @property
    def skew(self):
        if self._skew is None:
            self._skew = _skew(self._incr_evals[None, :])[0]
        return self._skew

    @property
    def quaternion(self):
        if self._quat is None:
            self._quat = _evecs_2_quat([self._evecs])[0]
        return Quaternion(self._quat.q)

    @property
    def spherical_repr(self):
        """
        Spherical representation of the tensor

        Returns a 3x3x3 array containing the isotropic, antisymmetric and
        symmetric parts of the tensor. The isotropic part is the average of
        the trace of the tensor, the antisymmetric part is the difference
        between the tensor and its transpose divided by 2, and the symmetric
        part is the sum of the tensor and its transpose divided by 2 minus
        the isotropic part. This construction is such that the sum of the
        components is equal to the original tensor.

        .. math::
            \\sigma = \\sigma_{iso} + \\sigma_{A} + \\sigma_{S}

        where

        .. math::
            \\sigma_{iso} = \\frac{1}{3} \\text{Tr}(\\sigma) \\mathbf{I}
                
            \\sigma_{A} = \\frac{1}{2} (\\sigma - \\sigma^T)

            \\sigma_{S} = \\frac{1}{2} (\\sigma + \\sigma^T) - \\sigma_{iso}

        Returns:
            np.ndarray -- 3x3x3 array containing the isotropic, antisymmetric
            and symmetric parts of the tensor
        
        """
        if self._sph is None:
            self._sph = np.zeros((3, 3, 3))
            # Isotropic part
            self._sph[0] = np.eye(3) * self.trace / 3
            # Anti-symmetric part
            self._sph[1] = (self._data - self._data.T) / 2.0
            # Symmetric part - isotropic part
            self._sph[2] = self._symm - self._sph[0]
            # _sph[0] + _sph[1] + _sph[2] == _data

        return self._sph.copy()

    def euler_angles(self, convention: str = "zyz", passive: bool = False, degrees=False) -> np.ndarray:
        """Euler angles of the Principal Axis System

        Return Euler angles of the PAS for this tensor in the
        required convention (currently supported: zyz, zxz).

        Args:
            convention {str} -- Euler angles convention to use
            (default: {'zyz'})
            passive {bool} -- Whether to return the passive Euler angles
            (default: {False})

        Returns:
            np.ndarray -- Euler angles in radians. Size of the array is (3,)
        """


        angles = _matrix_to_euler(self.eigenvectors, convention, passive)

        angles = _handle_euler_edge_cases(
                    angles,
                    self.eigenvalues,
                    self._symm,
                    convention = convention,
                    passive = passive
                    )
        # finally, test that the angles give a consistent rotation
        consistent_rotation = _test_euler_rotation(
                                    angles,
                                    self.eigenvalues,
                                    self.eigenvectors,
                                    convention,
                                    passive)
        if not consistent_rotation:
            warnings.warn("The Euler angles do not give a consistent rotation. "
                            "This is likely due to a degeneracy in the tensor. "
                            "Care must be taken when comparing the Euler angles of degenerate tensors.")

            # Re-running the Euler angle calculation with -self.eigenvectors
            self._evecs = -self.eigenvectors
            angles = _matrix_to_euler(self.eigenvectors, convention, passive)
            angles = _handle_euler_edge_cases(
                        angles,
                        self.eigenvalues,
                        self._symm,
                        convention = convention,
                        passive = passive
                        )

            # check again
            consistent_rotation = _test_euler_rotation(
                                    angles,
                                    self.eigenvalues,
                                    self.eigenvectors,
                                    convention,
                                    passive)
            # TODO raise error if still not consistent
            # if not consistent_rotation:
            #     raise ValueError("The Euler angles do not give a consistent rotation. "
            #                     "This is likely due to a degeneracy in the tensor. "
            #                     "Care must be taken when comparing the Euler angles of degenerate tensors.")

        if degrees:
            angles = np.degrees(angles)

        return angles

    def equivalent_euler_angles(self, convention="zyz", passive=False):
        '''
        Returns the equivalent Euler angles of the tensor.

        Args:
            convention {str} -- Euler angles convention to use
            (default: {'zyz'})
            passive {bool} -- Whether to return the passive Euler angles
            (default: {False})
        Returns:
            np.array -- Euler angles in radians. 
                        Size of the array is (4, 3) as there are 4 equivalent sets of Euler angles.


        '''
        euler_angles = self.euler_angles(convention, passive)

        return _equivalent_euler(euler_angles, passive=passive)

    def rotation_to(self, other):
        '''
        Returns the rotation matrix that rotates the tensor to the other tensor.

        '''
        R1 = self.eigenvectors
        R2 = other.eigenvectors
        R = np.linalg.inv(R1) @ R2
        # Need to guarantee that R expresses a *proper* rotation
        # (i.e. det(R) = 1)
        if np.linalg.det(R) < 0:
            R[:, 2] *= -1

        return R

    def euler_to(self, other, convention="zyz", passive=False, eps = 1e-6):
        '''
        Returns the Euler angles that rotate the tensor to the other tensor.
        '''
        convention = convention.lower()
        # first make sure they're not the same!
        if np.allclose(self.eigenvalues, other.eigenvalues):
            # check eigenvectors are the same up to a sign
            if np.allclose(self.eigenvectors, other.eigenvectors) or np.allclose(self.eigenvectors, -other.eigenvectors):
                warnings.warn("The tensors are identical. Returning zero Euler angles.")
                return np.zeros(3)
        if self.degeneracy == 2 or other.degeneracy == 2:
            Aevals = self.eigenvalues
            Bevals = other.eigenvalues
            # B (other) in the reference frame of A (self) - from paper
            # B_at_A = np.linalg.inv(self._symm) @ other._symm

            # alternative from the code
            R_A = Rotation.from_matrix(self.eigenvectors).as_matrix() # normalises the rotation matrix
            R_B = Rotation.from_matrix(other.eigenvectors).as_matrix() # normalises the rotation matrix
            Rrel1 = np.linalg.inv(R_B) @ R_A
            Rrel2 = np.linalg.inv(R_A) @ R_B

            B_at_A = Rrel1.dot(np.diag(Aevals).dot(np.linalg.inv(Rrel1)))

            if self.degeneracy == 2 and other.degeneracy == 2:
                # Both are axially symmetric - need to be careful

                # If both are axially symmetric, then do the following:
                # quick check if the angles are all zero:
                if np.abs(B_at_A[1,2] + B_at_A[0,2] + B_at_A[0,1]) < eps:
                    warnings.warn("The tensors are perfectly aligned. Returning zero Euler angles.")
                    return np.zeros(3)

                if convention == 'zyz':
                    # If both are axially symmetric, then
                    alpha = 0
                    gamma = 0
                    beta = np.arcsin(np.sqrt(
                        (B_at_A[2, 2]  - Aevals[2]) /
                        (Aevals[0] - Aevals[2])
                        ))
                    beta = np.abs(beta) # we can choose the sign of beta arbitrarily
                    return np.array([alpha, beta, gamma])

                if convention == 'zxz':
                    # in this convention, it depends on the unique axis
                    # of the other tensor
                    # but the possible angles are:
                    a = np.pi / 2
                    b = np.arcsin(-1 * np.sqrt(
                                (B_at_A[2, 2]  - Aevals[2]) /
                                (Aevals[0] - Aevals[2])
                                ))
                    b = np.abs(b) # we can choose the sign arbitrarily
                    c = 0


                    if np.abs(Bevals[0] - Bevals[1]) < eps:
                        # Unique axis is z
                        return np.array([a, b, c]) # 90, arcsin(...), 0
                    elif np.abs(Bevals[1] - Bevals[2]) < eps:
                        # Unique axis is x
                        return np.array([c, a, b]) # 0, 90, arcsin(...)
                    else:
                        raise ValueError('Unexpected eigenvalue ordering for axially symmetric tensor'
                                            'in zxz convention. Eigenvalues are: ', Bevals)
            elif self.degeneracy == 2 and other.degeneracy == 1:
                # If self is axially symmetric, but other isn't
                # just get the angles from Rrel1
                angles = _matrix_to_euler(Rrel1, convention, False) # always active here!
                angles = _handle_euler_edge_cases(
                    angles,
                    Aevals,
                    self._symm,
                    convention = convention,
                    passive = passive
                )
                if passive:
                    angles = angles[::-1]
                return angles
            elif self.degeneracy == 1 and other.degeneracy == 2:
                # If other is axially symmetric, but self isn't
                # TODO: check if we need to re-order Rrel2 to match ordering convention chosen for other tensor
                angles = _matrix_to_euler(Rrel2, convention, True) # always passive here!
                angles = _tryallanglestest(angles, np.diag(Aevals), np.diag(Bevals), Rrel1, convention)
                angles = -1*_handle_euler_edge_cases(
                    -1*angles[::-1],
                    Bevals,
                    other._symm,
                    convention = convention,
                    passive = passive
                )
                if not passive:
                    angles= angles[::-1]

                return angles





            if convention not in ["zyz", "zxz"]:
                # if neither zyz nor zxz, warn
                warnings.warn('Euler angles for axially symmetric tensors are only corrected for zyz and zxz conventions.'
                                ' Returning the uncorrected Euler angles.')

        if other.degeneracy == 2 and self.degeneracy != 2:
            # then let's swap them around, then transpose the result
            # TODO: test this!
            R = other.rotation_to(self).T
        else:
            R = self.rotation_to(other)
        return _matrix_to_euler(R.T, convention, passive)

    def equivalent_euler_to(self, other, convention="zyz", passive=False):
        '''
        Returns the equivalent Euler angles that rotate the tensor to the other tensor.
        '''
        euler_angles = self.euler_to(other, convention, passive)

        return _equivalent_relative_euler(euler_angles, passive=passive)

    @staticmethod
    def make_dipolar(
        a,
        i,
        j,
        cell=[0, 0, 0],
        isotopes={},
        isotope_i=None,
        isotope_j=None,
        rotation_axis=None,
    ):
        """Create a dipolar NMRTensor

        Create a dipolar NMR tensor from an atoms object and the indices
        of two atoms. Values are in Hz.

        | Args:
        |   a (ase.Atoms):      Atoms object of the structure to compute the
        |                       tensor for
        |   i (int):            index of first atom
        |   j (int):            index of second atom
        |   cell (np.array):    vector of the cell of the second atom, for
        |                       couplings between atoms in different cells.
        |                       By default is [0,0,0].
        |   isotopes (dict): dictionary of specific isotopes to use, by element
        |                    symbol. If the isotope doesn't exist an error will
        |                    be raised.
        |   isotope_i (int): isotope of atom i. To be used if
        |                    different atoms of the same element are supposed
        |                    to be of different isotopes. If None it will fall
        |                    back on the previous definitions. Otherwise it
        |                    overrides everything else.
        |   isotope_j (int): isotope of atom j. See above.
        |   rotation_axis (np.array):   an axis around which the selected pair
        |                               is rotating. If present, the tensor
        |                               will be averaged for infinitely fast
        |                               rotation around it.

        | Returns:
        |   diptens (NMRTensor):    an NMRTensor object with the dipolar
        |                           coupling matrix as data.

        """

        pos = a.get_positions()
        elems = np.array(a.get_chemical_symbols())
        r = pos[j] - pos[i] + np.dot(a.get_cell(), cell)
        rnorm = np.linalg.norm(r)
        gammas = _get_isotope_data(
            elems[[i, j]],
            "gamma",
            isotopes=isotopes,
            isotope_list=[isotope_i, isotope_j],
        )

        d = _dip_constant(rnorm * 1e-10, *gammas)

        evals = np.zeros(3)
        evecs = np.zeros((3, 3))

        if rotation_axis is None:
            axis = r / rnorm
        else:
            axis = np.array(rotation_axis) / np.linalg.norm(rotation_axis)
            vp2 = np.dot(r / rnorm, axis) ** 2
            d = 0.5 * d * (3 * vp2 - 1)

        evals[2] = 2 * d
        evals[:2] = -d

        evecs[:, 2] = axis

        x, y = axis[:2]
        evecs[1, 0] = x
        evecs[0, 0] = -y
        evecs[:, 0] /= (x ** 2 + y ** 2) ** 0.5

        evecs[0, 1]
        evecs[0, 1] = -evecs[2, 2] * evecs[1, 0]
        evecs[1, 1] = evecs[2, 2] * evecs[0, 0]
        evecs[2, 1] = evecs[0, 2] * evecs[1, 0] - evecs[1, 2] * evecs[0, 0]

        if d < 0:
            evals = evals[::-1]
            evecs = evecs[:, ::-1]

        D = [evals, evecs]

        return NMRTensor(D)

    def __repr__(self) -> str:
        """
        Return a string representation of the tensor
        """
        return f"NMRTensor(data={self.data})"

    def __str__(self) -> str:
        """
        Return a string representation of the tensor. Nicely print out the 3x3 matrix (data), the conventions, and the derived properties.
        """
        return f"NMRTensor with data: \n{self.data}\n\n" + \
                f"Isotropy: {self.isotropy:.5f}\n" + \
                f"Anisotropy: {self.anisotropy:.5f}\n" + \
                f"Asymmetry: {self.asymmetry:.5f}\n" + \
                f"Span: {self.span:.5f}\n" + \
                f"Skew: {self.skew:.5f}\n" + \
                f"Eigenvalues: {self.eigenvalues}\n" + \
                f"Eigenvectors: \n{self.eigenvectors}\n" + \
                f"Euler angles (deg): {self.euler_angles(degrees=True)}\n"

    def __array__(self, dtype=None):
        """
        Return a numpy array representation of the tensor.
        Required for NDArrayOperatorsMixin.
        """
        return np.asarray(self._data, dtype=dtype)

    @property
    def _initialisation_params(self):
        """
        Return the parameters used for the initialisation of the tensor.
        This is useful for debugging and understanding how the tensor was created.
        """
        return {
            "order": self.order
        }

    def _create_like(self, data):
        """
        Create a new tensor of the same type with the provided data
        but preserving other properties of this tensor.
        
        Args:
            data (np.ndarray): New tensor data
            
        Returns:
            NMRTensor: A new tensor with the same properties
        """
        params = self._initialisation_params
        return self.__class__(data = data, **params)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """
        Handle NumPy's universal functions for tensor operations.
        This enables proper handling of arithmetic operations.
        """
        # Only handle the __call__ method of ufuncs
        if method != '__call__':
            return NotImplemented

        # Check if the operation is supported
        supported_ops = [np.add, np.subtract, np.multiply, np.true_divide, np.matmul, np.negative, np.positive]
        if ufunc not in supported_ops:
            return NotImplemented

        # Handle unary operations
        if len(inputs) == 1:
            if ufunc is np.negative or ufunc is np.positive:
                # Let subclasses handle their own instantiation if possible
                data = -self.data if ufunc is np.negative else self.data
                return self._create_like(data)
            return NotImplemented

        # Handle binary operations
        if len(inputs) == 2:
            # Find which input is self
            self_idx = None
            for i, x in enumerate(inputs):
                if isinstance(x, NMRTensor) and x is self:
                    self_idx = i
                    break

            if self_idx is None:
                return NotImplemented

            # Get the other operand
            other_idx = 1 - self_idx
            other = inputs[other_idx]

            # Convert inputs to numpy arrays
            tensor_data = self.data

            # Process the other operand based on its type
            if isinstance(other, NMRTensor):
                # Check parameter consistency if both are same subclass
                if isinstance(other, self.__class__):
                    self._check_binary_op_params(other)
                # Check if tensors have the same order, warn if not
                if other.order != self.order:
                    warnings.warn(f"Operating on tensors with different ordering conventions: "
                                f"{self.order} and {other.order}. Using {self.order}.")
                other_data = other.data
            elif isinstance(other, (int, float)):
                other_data = other
            elif isinstance(other, np.ndarray):
                # Special handling for matmul with vectors
                if ufunc is np.matmul and len(other.shape) == 1 and other.shape[0] == 3 or other.shape == (3, 3):
                    other_data = other
                else:
                    raise ValueError(f"Cannot perform operation with array of shape {other.shape}")
            else:
                return NotImplemented

            # Apply the operation with correct argument order
            args = (tensor_data, other_data) if self_idx == 0 else (other_data, tensor_data)
            result_data = ufunc(*args, **kwargs)

            # Return appropriate result based on the operation
            if ufunc in [np.add, np.subtract, np.multiply, np.true_divide]:
                return self._create_like(result_data)
            elif ufunc is np.matmul:
                if result_data.shape == (3, 3):
                    return self._create_like(result_data)
                else:
                    return result_data

        # For unsupported operations
        return NotImplemented

    def _check_binary_op_params(self, other):
        """
        Check that parameters of two tensors are compatible for binary operations.
        Warns about any inconsistencies.

        Args:
            other (NMRTensor): Another tensor to compare parameters with
        """
        # Convenience method to check parameters for two tensors
        self._check_compatible([self, other])


    @classmethod
    def _check_compatible(cls, tensors: list[T]) -> None:
        """
        Tensor compatibility checks.

        Args:
            tensors (list[T]): list of NMRTensor objects

        Raises:
            ValueError: If the tensors are not compatible
        """
        first_tensor = tensors[0]
        expected_params = first_tensor._initialisation_params

        param_keys = expected_params.keys()

        for t in tensors[1:]:
            tensor_params = t._initialisation_params
            for key in param_keys:
                if tensor_params.get(key) != expected_params.get(key):
                    raise ValueError(f"All tensors must have the same {key}. "
                                    f"Found: {[t._initialisation_params.get(key) for t in tensors]}")


    @classmethod
    def mean(cls: type[T], tensor_list: list[Any], axis: Optional[int]=None, weights: Optional[np.ndarray]=None) -> Union[T, list[T]]:
        """
        Calculate the mean of a list of NMRTensor objects along a specified axis.

        Args:
            tensor_list (list[Any]): list or nested list of NMRTensor objects to average.
                For a 2D array with shape [N, M], this represents N rows of M tensors.
            axis (int, optional): Axis along which the mean is computed:
                - None: Average all tensors into a single tensor
                - 0: Average over rows (result has length M)
                - 1: Average over columns (result has length N)
                - etc. for higher dimensions
            weights (np.ndarray, optional): Array of weights to use for weighted average.
                Must have length matching the dimension being averaged:
                - If axis=None: weights should have same length as flattened tensor_list
                - If axis=0: weights should have length N (one weight per row)
                - If axis=1: weights should have length M (one weight per column)

        Returns:
            Union[T, list[T]]: A new tensor or list of tensors with averaged values

        Raises:
            ValueError: If tensor_list is empty, has incompatible dimensions,
                       or if weights have incorrect shape
        """
        if not tensor_list:
            raise ValueError("Cannot calculate mean of an empty list")

        # Helper function for averaging a flat list of tensors
        def average_tensor_list(tensors, weights=None):
            if not tensors:
                raise ValueError("Cannot calculate mean of an empty list")

            try:
                cls._check_compatible(tensors)
            except ValueError as e:
                raise ValueError("It seems you are trying to average incompatible tensors. \n"
                                 f"{e}\n"
                                 " Did you mean to average over a different axis perhaps?\n"
                                 " If so, please use the axis argument (e.g. axis=0).\n")
            tensor_data = np.array([t.data for t in tensors])

            if weights is not None:
                if len(weights) != len(tensors):
                    raise ValueError(f"Weights length ({len(weights)}) doesn't match tensor count ({len(tensors)})")
                mean_data = np.average(tensor_data, axis=0, weights=weights)
            else:
                mean_data = np.mean(tensor_data, axis=0)

            return tensors[0]._create_like(mean_data)

        # Handle 1D list of tensors with axis=None
        if all(isinstance(t, NMRTensor) for t in tensor_list) and axis is None:
            return average_tensor_list(tensor_list, weights)

        # Convert to numpy object array for easier handling of dimensions
        tensor_array = np.array(tensor_list, dtype=object)
        array_shape = tensor_array.shape[:-2]  # Exclude the last two dimensions (3x3)

        # For axis=None, flatten the array and average all tensors
        if axis is None:
            flat_tensors = tensor_array.flatten().tolist()
            if weights is not None and len(weights) != len(flat_tensors):
                raise ValueError(f"For axis=None, weights length ({len(weights)}) must match "
                                 f"total tensor count ({len(flat_tensors)})")
            return average_tensor_list(flat_tensors, weights)

        # Validate axis
        if axis >= len(array_shape):
            raise ValueError(f"Axis {axis} out of bounds for array of dimension {len(array_shape)}")

        # Validate weights for the specified axis
        if weights is not None:
            if len(weights) != array_shape[axis]:
                raise ValueError(f"For axis={axis}, weights length ({len(weights)}) must match "
                                 f"dimension size ({array_shape[axis]})")

        # Use numpy's apply_along_axis to handle the averaging
        result_shape = list(array_shape)
        result_shape.pop(axis)  # Remove the dimension being averaged
        result = np.empty(result_shape, dtype=object)

        # Helper function to apply averaging along the specified axis
        def average_along_axis():
            # For each slice perpendicular to the specified axis
            for idx in np.ndindex(result_shape):
                idx_list = list(idx)
                if axis > 0:
                    idx_list.insert(axis, slice(None))
                else:
                    idx_list = [slice(None)] + idx_list

                # Extract tensors along the axis
                tensors_to_average = tensor_array[tuple(idx_list)].tolist()

                # Apply weights if provided
                if weights is not None:
                    avg_tensor = average_tensor_list(tensors_to_average, weights)
                else:
                    avg_tensor = average_tensor_list(tensors_to_average)

                result[idx] = avg_tensor
            return result

        # Handle 1D and 2D cases explicitly for clarity
        if len(array_shape) == 1:
            # Simple 1D case - single average
            return average_tensor_list(tensor_list, weights)
        elif len(array_shape) == 2:
            # Common 2D case - handle more explicitly
            if axis == 0:  # Average over rows
                result = np.empty(array_shape[1], dtype=object)
                for col in range(array_shape[1]):
                    # Extract all tensor objects from tensor_list in the column
                    col_tensors = [row[col] for row in tensor_list]
                    result[col] = average_tensor_list(col_tensors, weights)
                return result.tolist()
            elif axis == 1:  # Average over columns
                result = np.empty(array_shape[0], dtype=object)
                for row in range(array_shape[0]):
                    row_tensors = tensor_list[row]
                    result[row] = average_tensor_list(row_tensors, weights)
                return result.tolist()

        # Generic case for higher dimensions
        return average_along_axis().tolist()

    def __eq__(self, other):
        """
        Check if tensor equals another tensor or array
        """
        if isinstance(other, NMRTensor):
            return np.allclose(self.data, other.data)
        elif isinstance(other, np.ndarray) and other.shape == (3, 3):
            return np.allclose(self.data, other)
        else:
            return False



class MagneticShielding(NMRTensor):
    """MagneticShielding

    Class containing a magnetic shielding tensor, a subclass of NMRTensor

    It provides easy access to common representations of the tensor in common conventions
    such as: IPAC, Herzfeld-Berger, Haeberlen, Maryland, and Mehring.
    For more information on these conventions, see the documentation for the corresponding
    NamedTuple and also :cite:p:`Harris2008`

    """

    # Define some NamedTuples for the different representations
    # of the tensor. See here for a description of the different
    # representations:
    # http://anorganik.uni-tuebingen.de/klaus/nmr/index.php?p=conventions/csa/csa
    class IUPACNotaion(NamedTuple):
        """
        IUPAC convention :cite:p:`Mason1993` (equivalent to the Mehring convention :cite:p:`Mehring1983`). 
        Follows the high frequency-positive order.
        Therefore:
        :math:`\\sigma_{11}` corresponds to the direction of least shielding, with the highest frequency,
        :math:`\\sigma_{33}` corresponds to the direction of highest shielding, with the lowest frequency.

        .. math::
            \\sigma_{11} \\leq \\sigma_{22} \\leq \\sigma_{33}

        The isotropic value, :math:`\\sigma_{iso}`, is the average values of the principal components,
        and corresponds to the center of gravity of the line shape.

        Note that the IUPAC convention is equivalent to the Mehring convention.

        """
        sigma_iso: float
        sigma_11: float
        sigma_22: float
        sigma_33: float

        def __str__(self):
            return (f"IUPACNotaion/MehringNotation:\n"
                    f"  sigma_11:  {self.sigma_11:.5f}\n"
                    f"  sigma_22:  {self.sigma_22:.5f}\n"
                    f"  sigma_33:  {self.sigma_33:.5f}\n"
                    f"  sigma_iso: {self.sigma_iso:.5f}")

    # Herzfeld-Berger Notation
    class HerzfeldBergerNotation(NamedTuple):
        """
        Herzfeld-Berger convention :cite:p:`Herzfeld1980` uses the following parameters:

        * :math:`\\sigma_{iso}` : isotropic magnetic shielding
        * :math:`\\Omega = \\sigma_{33} - \\sigma_{11}` (where these are the max and min principal components respectively): the span 
        * :math:`\\kappa = 3(\\sigma_{iso} - \\sigma_{22}) / \\Omega` (where :math:`\\sigma_{22}` is the median principal component): the skew

        """
        sigma_iso: float
        omega: float
        kappa: float

        def __str__(self):
            return (f"HerzfeldBergerNotation/MarylandNotation:\n"
                    f"  sigma_iso (isotropy): {self.sigma_iso:.5f}\n"
                    f"  omega (span):     {self.omega:.5f}\n"
                    f"  kappa (skew):     {self.kappa:.5f}")
    class MarylandNotation(HerzfeldBergerNotation):
        """The same as the Herzfeld-Berger notation."""

    class HaeberlenNotation(NamedTuple):
        """
        Haeberlen convention :cite:p:`Haeberlen1976` follows this ordering:
        
        .. math::
            |\\sigma_{zz} - \\sigma_{iso} | \\geq |\\sigma_{xx} - \\sigma_{iso} | \\geq |\\sigma_{yy} - \\sigma_{iso} |


        and uses the following parameters to describe the shielding tensor:

        * :math:`\\sigma_{iso}` : isotropic magnetic shielding
        * :math:`\\sigma = \\sigma_{zz} - \\sigma_{iso}` : the reduced anisotropy
        * :math:`\\Delta = \\sigma_{zz} - (\\sigma_{xx} + \\sigma_{yy}) / 2 = 3\\sigma / 2`  : the anisotropy
        * :math:`\\eta = (\\sigma_{yy} - \\sigma_{xx}) / \\sigma` : the asymmetry ( :math:`0 \\leq \\eta \\leq +1` )

        """

        sigma_iso: float
        sigma: float
        delta: float
        eta: float

        def __str__(self):
            return (f"HaeberlenNotation:\n"
                    f"  sigma_iso (isotropy): {self.sigma_iso:.5f}\n"
                    f"  sigma (reduced anisotropy): {self.sigma:.5f}\n"
                    f"  delta (anisotropy): {self.delta:.5f}\n"
                    f"  eta (asymmetry): {self.eta:.5f}")

    class MehringNotation(IUPACNotaion):
        """
        Mehring convention :cite:p:`Mehring1983` is equivalent to the IUPAC convention.
        """


    def __init__(self,
        data: Union[np.ndarray, tuple[np.ndarray, np.ndarray]],
        species: str,
        order:str = NMRTensor.ORDER_HAEBERLEN,
        reference:Optional[float]=None,
        gradient: float=-1.0,
        tag = None):
        """
        Initialise the MSTensor

        Create an MSTensor object from a 3x3 matrix.

        Arguments:
            species (str):      Element or isotope symbol of the atom the tensor refers to. e.g 'H', 'C', '13C'
            data (np.ndarray or tuple):  3x3 matrix containing the tensor, or
                                         pair [evals, evecs] for the symmetric
                                         part alone.
            order (str):        Order to use for eigenvalues/eigenvectors. Can
                                be 'i' (ORDER_INCREASING), 'd'
                                (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                                'n' (ORDER_NQR). Default is 'h' for MS tensors.
            references (dict):  Dictionary of references for the magnetic shielding 
                                to chemical shift conversion. Keys are element symbols,
                                values are the reference chemical shift in ppm.
            gradients (dict):   Dictionary of gradients for the magnetic shielding
                                tensor. Keys are element symbols, values are the
                                gradient. Any unspecified gradients will be set to -1.
            tag (str):          Optional tag to identify the tensor. In a magres file, this would be the
                                'ms_sometag' though for most magres file ms is not decomposed into contributions and the tag
                                is simply 'ms'. By default, this is None.


        """
        super().__init__(data, order=order)
        self.species = species
        self.reference = reference
        self.gradient = gradient
        self.mstag = tag

    @property
    def element(self):
        '''
        Returns the element of the tensor. Species could have the isotope info, but
        here we just want the element.
        e.g. 13C -> C
        1H -> H
        '''
        return self.species.strip('1234567890')

    @property
    def shift(self):
        '''
        Returns the isotropic chemical shift of the tensor (ppm).
        If the reference is not set, will raise a ValueError. You can set the reference
        using the reference attribute.
        '''
        if self.reference is None:
            raise ValueError('Reference chemical shift not set for this tensor.')

        return self.reference + (self.gradient * self.isotropy) / (1 + self.reference * 1e-6)

    @property
    def shift_tensor(self):
        '''
        Returns the chemical shift tensor in ppm.
        '''
        if self.reference is None:
            raise ValueError('Reference chemical shift not set for this tensor.')
        shift_data = self.reference + (self.gradient * self.data) / (1 + self.reference * 1e-6)
        return MagneticShielding(shift_data, self.species, order=self.order, reference=self.reference, gradient=self.gradient)

    @property
    def haeberlen_values(self):
        """The magnetic shielding tensor in Haeberlen Notation."""
        sigma_iso = self.isotropy
        sigma = self.reduced_anisotropy
        delta = self.anisotropy
        eta = self.asymmetry
        return self.HaeberlenNotation(sigma_iso, sigma, delta, eta)

    @property
    def mehring_values(self):
        """The magnetic shielding tensor in Mehring Notation."""
        sigma_iso = self.isotropy
        # sort the eigenvalues in increasing order
        sigma_11, sigma_22, sigma_33 = sorted(self.eigenvalues)
        return self.MehringNotation(sigma_iso, sigma_11, sigma_22, sigma_33)

    @property
    def iupac_values(self):
        """The magnetic shielding tensor in IUPAC Notation."""
        sigma_iso = self.isotropy
        # sort the eigenvalues in increasing order
        sigma_11, sigma_22, sigma_33 = sorted(self.eigenvalues)
        return self.IUPACNotaion(sigma_iso, sigma_11, sigma_22, sigma_33)

    @property
    def maryland_values(self):
        """The magnetic shielding tensor in Maryland Notation."""
        sigma_iso = self.isotropy
        omega = self.span
        kappa = self.skew
        return self.MarylandNotation(sigma_iso, omega, kappa)
    @property
    def herzfeldberger_values(self):
        """The magnetic shielding tensor in Herzfeld-Berger Notation."""
        sigma_iso = self.isotropy
        omega = self.span
        kappa = self.skew
        return self.HerzfeldBergerNotation(sigma_iso, omega, kappa)

    # Set/update the reference chemical shift
    def set_reference(self, reference: float):
        '''
        Set the reference chemical shift for this tensor.
        '''
        self.reference = reference

    # Set/update the gradient
    def set_gradient(self, gradient: float):
        '''
        Set the gradient for this tensor.
        '''
        self.gradient = gradient

    @property
    def _initialisation_params(self):
        """
        Return the parameters used for initialisation of the tensor.
        """
        return {
            'species': self.species,
            'order': self.order,
            'reference': self.reference,
            'gradient': self.gradient,
            'tag': self.mstag
        }

    def __str__(self):
        """
        Neatly formatted string representation of the tensor and Haeberlen description.
        """
        s =  f"Magnetic Shielding Tensor for {self.species}:\n"
        # 3x3 tensor representation neatly formatted with 5 decimal places
        s += str(np.array2string(self.data, precision=5, separator=",", suppress_small=True)) + "\n"
        s += str(self.haeberlen_values)
        return s

class ElectricFieldGradient(NMRTensor):
    """ElectricFieldGradient

    Class containing an electric field gradient tensor, a subclass of NMRTensor.

    The EFG tensor is a symmetric 3x3 matrix, with a trace of zero and has therefore
    only 5 independent degrees of freedom. These are often captured using the following: 

    * largest absolute eigenvalue (:math:`V_{zz}`)
    * the asymmetry parameter ( :math:`\\eta` )
    *  α, β, γ Euler angles describing the orientation of the tensor.

    The principal components of the tensor are the eigenvalues of the tensor, and they
    are usually sorted according to the NQR convention ( :math:`|V_{zz}| \\geq |V_{yy}| \\geq |V_{xx}|` ). Note however, that 
    Simpson uses the convention :math:`|V_{zz}| \\geq |V_{xx}| \\geq |V_{yy}|` . This is can be compensated for by using the
    order parameter when creating the tensor.

    Note that some conventions use the reduced anisotropy ( :math:`\\zeta` ) instead of the asymmetry parameter.

    """

    def __init__(self,
                data: Union[np.ndarray, tuple[np.ndarray, np.ndarray]],
                species: str,
                order: str=NMRTensor.ORDER_NQR,
                quadrupole_moment:Optional[float] = None,
                gamma:Optional[float] = None):
        """
        Initialise the EFGTensor

        Create an EFGTensor object from a 3x3 matrix.

        Args:
            data (np.ndarray or tuple):  3x3 matrix containing the tensor, or
                                         pair [evals, evecs] for the symmetric
                                         part alone.
            species (str):      Isotope symbol of the atom the tensor refers to. e.g '2H', '13C'
            order (str):        Order to use for eigenvalues/eigenvectors. Can
                                be 'i' (ORDER_INCREASING), 'd'
                                (ORDER_DECREASING), 'h' (ORDER_HAEBERLEN) or
                                'n' (ORDER_NQR). Default is 'h' for EFG tensors.
            quadrupole_moment (float): Quadrupole moment of the nucleus in millibarn.
            gamma (float):      Nuclear gyromagnetic ratio in rad/(s*T). If not provided, will be
                                looked up using the species.
        """
        super().__init__(data, order=order)
        self.species = species
        isotope_number, element = _split_species(species)
        self.element = element

        # Nuclear spin in Bohr magnetons
        self.spin = nmr_spin(element, iso=isotope_number)

        # Quadrupole moment in barns
        self.quadrupole_moment = quadrupole_moment or nmr_quadrupole(element, iso=isotope_number)

        # gamma in rad/(s*T)
        self.gamma = gamma or nmr_gamma(element, iso=isotope_number)



    @property
    def eta(self):
        '''
        Returns the asymmetry parameter of the tensor.
        '''
        return self.asymmetry

    @property
    def zeta(self):
        '''
        Returns the reduced anisotropy of the tensor.
        '''
        return self.reduced_anisotropy

    @property
    def Vzz(self):
        '''
        Returns the largest absolute eigenvalue of the tensor (
        the principal component of the EFG tensor).
        This should be in atomic units (a.u.) if read in from e.g. a magres file.
        '''
        # The eigenvalues are already sorted in abs increasing order (NQR convention)
        if self.order == self.ORDER_NQR:
            return self.eigenvalues[2]
        else:
            # warn user that the eigenvalues are not sorted in NQR order
            warnings.warn("The eigenvalues are not sorted in NQR order. "
                            "Returning the largest abs. eigenvalue following NQR convention.")
            # sort them in NQR order
            return _evals_sort(self.eigenvalues, convention=self.ORDER_NQR)[2]

    @property
    def Cq(self) -> float:
        '''
        Calculates the quadrupolar constant in Hz for this EFG tensor.
        The constant will be zero for non-quadrupole active nuclei.
        The quadrupole moment used is that for the nucleus of the isotope
        specified in the species attribute.

        This property is defined as

        .. math::

            C_Q = \\frac{e^2qQ}{h}

        in Hz. It is important to keep in mind that therefore this represents a
        *frequency*; the corresponding 'omega' (pulsation) would be the same value
        multiplied by 2*pi. This is, for example, exactly the value required as
        input in Simpson's SPINSYS section.

        Returns:
            float: Quadrupolar constant value in Hz.
        '''

        return EFG_TO_CHI * self.quadrupole_moment * self.Vzz

    @property
    def Pq(self) -> float:
        '''
        Calculates the quadrupolar product in Hz for this EFG tensor.

        .. math::

            P_Q = C_Q (1+\\frac{\\eta_Q^2}{3})^{1/2}

        Returns:
            float: Quadrupolar product value in Hz.
        '''
        return self.Cq * (1 + self.eta ** 2 / 3)**0.5

    @property
    def nuq(self) -> float:
        '''
        Calculates the quadrupolar frequency in Hz for this EFG tensor.
        This is also known as the quadrupolar splitting parameter.

        .. math::

            \\nu_Q = \\frac{3C_{Q}}{2I(2I-1)}

        where I is the nuclear spin of the nucleus.

        See this for conventions:
        http://anorganik.uni-tuebingen.de/klaus/nmr/index.php?p=conventions/efg/quadtools

        Returns:
            float: Quadrupolar frequency value in Hz.
        '''
        return 3 * self.Cq / (2 * self.spin * (2 * self.spin - 1))

    def get_larmor_frequency(self, Bext):
        '''
        Returns the Larmor frequency of the nucleus in an external magnetic field in Hz.

        .. math::
            
                \\nu_L = \\gamma B_{ext} / (2\\pi)

        where :math:`\\gamma` is the gyromagnetic ratio of the nucleus in rad/(s*T) and :math:`B_{ext}` is the external magnetic field in T.

        Args:
            Bext (float): External magnetic field in T.

        Returns:
            float: Larmor frequency in Hz.
        '''
        return self.gamma * Bext / (2 * np.pi)


    def get_quadrupolar_perturbation(self, Bext):
        '''
        Returns the perturbation of the quadrupolar Hamiltonian due to an external magnetic field.
        The perturbation is given by:

        .. math::

            a = \\frac{\\nu_Q^2}{\\nu_L} (I(I+1) - 3/2)

        where :math:`\\nu_Q` is the quadrupolar frequency and :math:`\\nu_L` is the Larmor frequency.

        Args:
            Bext (float): External magnetic field in T.

        Returns:
            float: Perturbation in Hz.
        '''
        nu_larmor = self.get_larmor_frequency(Bext)
        spin = self.spin
        nuq = self.nuq
        a = (nuq**2 / nu_larmor) * (spin*(spin+1) - 3/2)
        return a

    def get_MAS_full_max_linewidth(self, Bext):
        '''
        Returns the MAS full maximum line width (Hz) of the central transition.

        Defined as:

        .. math::
            
                \\Delta\\nu = a\\frac{(6+\\eta)^2}{504}

        where :math:`a` is the quadrupolar perturbation and :math:`\\eta` is the asymmetry parameter.

        Args:
            Bext (float): External magnetic field in T.

        Returns:
            float: MAS full maximum line width in Hz.
        '''
        a = self.get_quadrupolar_perturbation(Bext)
        return a * (6 + self.eta)**2 / 504

    @property
    def NQR(self):
        '''
        EFGNQR

        Produces an array containing NQR transition frequencies (in Hz) for every atom
        in a system. For non-quadrupole active nuclei, the we return an empty dictionary.
        Unless specified otherwise, the spin and quadrupole moment of the most common
        NMR-active isotope is used.

        For reference: the value returned by this property is defined as

        .. math::

            A = \\frac{V_{zz} Q}{4I(2I - 1)}
            fq = 3A(2m+1)\\sqrt{1 + \\eta^2/3}

        in Hz.
        It is important to keep in mind that therefore this represents a
        *frequency*; the corresponding 'omega' (pulsation) would be the same value
        multiplied by 2*pi. 
        '''
        result = defaultdict(None)

        if self.spin == 0:
            return result

        A = EFG_TO_CHI * self.Vzz * self.quadrupole_moment / (4 * self.spin * (2 * self.spin - 1))
        ms = [m for m in _frange(-self.spin, self.spin + 1, 1) if m >= 0.0][:-1]
        for m in ms:
            key = f'm={m}->{m+1}'
            fq = 3 * A * (2 * m + 1) * np.sqrt(1 + self.eta ** 2 / 3)
            result[key] = fq

        return result


    # def get_MAS_second_order_shift(self, Bext):
    #     '''
    #     Returns the MAS second order shift (ppm) of the central transition.

    #     TODO: eq. taken from here: http://anorganik.uni-tuebingen.de/klaus/nmr/index.php?p=conventions/efg/quadtools
    #     double-check.

    #     Defined as:

    #     .. math::

    #             \\nu_{c.g.} = \\nu_{0} - \\frac{1}{30} a (1 + \\eta^{\\frac{2}{3}})

    #     Args:
    #         Bext (float): External magnetic field in T.

    #     Returns:
    #         float: MAS second order shift in ppm.
    #     '''
    #     a = self.get_quadrupolar_perturbation(Bext)
    #     nu_larmor = self.get_larmor_frequency(Bext)
    #     return nu_larmor - (a / 30) * (1 + self.eta**(2/3))

    @property
    def _initialisation_params(self):
        """
        Return the parameters used for initialisation of the tensor.
        """
        return {
            'species': self.species,
            'order': self.order,
            'quadrupole_moment': self.quadrupole_moment,
            'gamma': self.gamma
        }



def contains_nmr_tensors(values):
    """Check if values contain NMRTensor objects at any nesting level"""
    if isinstance(values, (list, np.ndarray)):
        if len(values) > 0:
            if isinstance(values[0], NMRTensor):
                return True
            elif isinstance(values[0], (list, np.ndarray)):
                return contains_nmr_tensors(values[0])
    return False



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" A script to extract NMR properties from a (list of) .magres files"""

# TODO:
# - Add support for more properties
# - Add selection by X
#    - got sumo-style atom selection down. What else would be good to implement now?
#  ~~Add support for custom isotopes~~
# - ~~Add function to symmetry-reduce the output -- merge symmetry-equivalent sites~~
#    - e.g. get asymmetric unit cell for molecular crystals



import argparse as ap
import os
import re
import sys
import warnings

import numpy as np
from ase import io

from soprano.properties.labeling import UniqueSites
from soprano.properties.nmr import *
from soprano.selection import AtomSelection

__author__ = "J. Kane Shenton"
__maintainer__ = "J. Kane Shenton"
__email__ = "kane.shenton@stfc.ac.uk"
__date__ = "March 21, 2022"



def isotope_selection(selection_string):
    """Parse the isotope string. Adapted from [sumo](https://github.com/SMTG-UCL/sumo)
    Args:
        selection_string (str): The isotopes specification, in the form ``"2H,15N" for deuterium and 15N``.
    Returns:
        dict: The isotope for each element specified. Formatted as::
            {Element: Isotope}.
    """
    isotope_dict = {}
    for split in selection_string.split(","):
        matchobj = re.match(r"(\d+)([A-Z][a-z]?)", split)
        el = matchobj.group(2)
        isotope = int(matchobj.group(1))
        isotope_dict[el] = isotope
    return isotope_dict

def __main__():

    parser = ap.ArgumentParser(
        description="""
        Processes .magres files containing NMR-related properties 
        and prints a summary. It defaults to printing all NMR properties 
        present in the file for all the atoms. 
        
        See the below arguments for how to extract specific information.
        """,
        epilog=f"""
        Author: {__author__} ({__email__})
        Last updated: {__date__}
        """
    )
    # Main argument
    parser.add_argument(
        "input_files",
        type=str,
        nargs="+",
        default=None,
        help="Magres files for which to extract the summary data. "
        "You can use standard shell wildcards such as *.magres",
    )
    # Optional arguments
    parser.add_argument(
        "-s",
        "--select",
        type=str,
        metavar="A",
        help=('element/atoms to include (e.g. "C" for only carbon or '
        '"C.1.2.3,H.1.2" for carbons 1,2,3 and hydrogens 1 and 2)'),
    )
    parser.add_argument(
        "-i",
        "--isotope",
        type=isotope_selection,
        metavar="I",
        default={},
        help=('Isotopes specification (e.g. "13C" for carbon 13 '
        '"2H,15N" for deuterium and N 15). '
        'When nothing is specified it defaults to the most common NMR active isotope.'),
    )
    # add argument to reduce by symmetry
    parser.add_argument(
        "-r",
        "--reduce",
        action="store_true",
        help=("Reduce the output by symmetry-equivalent sites. "
        "Note that this doesn't take into account magnetic symmetry!"),
    )
    #add option to average over reduced groups
    parser.add_argument(
        "-a",
        "--average",
        action="store_true",
        help=("Average over symmetry-equivalent sites. "
        "Note that this doesn't take into account magnetic symmetry!"),
    )
    # add option to set the symprec
    parser.add_argument(
        "--symprec",
        type=float,
        default=1e-4,
        help=("Set the symprec for the symmetry reduction. "
        "Default is 1e-4."),
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action='store_true',
        default=False,
        help="If present, suppress print headers and append filename to each line",
    )
    parser.add_argument(
        "-p",
        "--properties",
        type=str,
        metavar="P",
        nargs="+",
        default=["ms", "efg"],
        help="Properties for which to extract and summarise e.g. '-p ms efg'"
         "(default: ms and efg)",
    )
    # optional argument for the precision of the output
    parser.add_argument(
        "--precision",
        type=int,
        default=3,
        help="Precision of the output (default: 3) -- number of decimal places",
    )
    ## add optional argument for euler angle convention
    parser.add_argument(
        "--euler",
        type=str,
        default="zyz",
        help="Euler angle convention (default: zyz). Can be either 'zyz' or 'zxz'",
    )

    ## output csv? default False
    parser.add_argument('--csv',    action='store_true',
        help="output data to a CSV file for each .magres file"
        "(default: False)")
    parser.add_argument('--no-csv', action='store_false')
    parser.set_defaults(csv=False)

    parser.add_argument(
        "--prefix", type=str, default="", help="Prefix added to CSV output file"
    )

    args = parser.parse_args()
    properties = [p.lower() for p in args.properties]


    # how many files are to be summarised?
    nfiles = len(args.input_files)

    # label precision
    PREC = args.precision

    # comma or tab separated?
    if args.csv:
        sep = ',   '
    else:
        sep = '\t'

    for f in args.input_files:
        dirname = os.path.dirname(f)
        basename = os.path.basename(f)


        MS_HEADER  = '\n'+'\n'.join([
            '#'+155*'-'+'#',
            '# {:^153} #'.format(f'Magnetic Shielding:  {f}') ,
            '#'+155*'-'+'#'])

        EFG_HEADER = '\n'+'\n'.join([
            '#'+155*'-'+'#',
            '# {:^153} #'.format(f'Electric Field Gradient:  {f}') ,
            '#'+155*'-'+'#'])

        # try to read in the file:
        try:
            atoms = io.read(f)
        except OSError:
            print(f"File {f} not found, skipping")
            continue


        # Do they actually have any magres data?
        if not any([atoms.has(k) for k in properties]):
            print("File {0} has no {1} data extract, skipping".format(f, ' '.join(properties)))
            continue

        # create new array for multiplicities
        multiplicities = np.ones(len(atoms))
        atoms.set_array('multiplicities', multiplicities)

        # reduce by symmetry?
        if args.reduce:
            symprec = args.symprec
            tags = UniqueSites.get(atoms, symprec=symprec)
            groups = [np.where(tags == i)[0] for i in range(max(tags))]

            # update multiplicities
            group_multiplicities = [len(g) for g in groups]
            for i, g in enumerate(groups):
                multiplicities[g] = group_multiplicities[i]
            atoms.set_array('multiplicities', multiplicities)
            # average properties within each group
            if args.average:
                for p in properties:
                    arr = atoms.get_array(p)
                    for group in groups:
                        # print(f"Average {p} over symmetry-equivalent sites {group}")
                        arr[group] = np.average(arr[group], axis=0)
                    atoms.set_array(p, arr)

            # create a new atoms object with only the unique sites
            # taking the first site from each group
            uniqueinds = [groups[i][0] for i in range(len(groups))]
            atoms = AtomSelection(atoms, uniqueinds).subset(atoms)


            print(atoms.get_array('multiplicities'))
            # atoms = AtomSelection.unique(atoms, symprec=symprec).subset(atoms)

        # select subset of atoms based on selection string
        if args.select:
            sel_selectionstring = AtomSelection.from_selection_string(atoms, args.select)
            atoms = sel_selectionstring.subset(atoms)

        # Note we could have combined selections as in Tutorial 3, but then
        # we lose the nice ordering of the atoms so better to apply selections successively...

        # Obtain labels for subset atoms:
        indices = atoms.get_array('indices')
        labels = atoms.get_array('labels')



        # --- ms --- #
        if 'ms' in properties:
            # list to hold the summary for this file
            summary = []
            try:
                # quick check that ms data is present
                mstensor = atoms.get_array('ms')

                # Isotropy, Anisotropy and Asymmetry (Haeberlen convention)
                iso   = MSIsotropy.get(atoms)
                aniso = MSAnisotropy.get(atoms)
                red_aniso = MSReducedAnisotropy.get(atoms)
                asymm = MSAsymmetry.get(atoms)
                # Span and skew
                span = MSSpan.get(atoms)
                skew = MSSkew.get(atoms)
                # quaternion
                quat = MSQuaternion.get(atoms)


                # if not quiet, print header
                if not args.quiet:
                    if not args.csv:
                        summary += [MS_HEADER.format(fname=f)]
                    table_headers = ['Label',
                                     'Index',
                                     'Isotropy/ppm',
                                     'Anisotropy/ppm',
                                     'Red.anisotropy/ppm',
                                     'Asymmetry',
                                     'Span/ppm',
                                     'Skew',
                                     'alpha/deg',
                                     'beta/deg',
                                     'gamma/deg']
                    summary += [f'{sep}'.join([f'{lab:<9}' for lab in table_headers])]
                for i, jl in enumerate(labels):
                    a, b, c = quat[i].euler_angles(mode=args.euler)*180/np.pi # rad to degrees
                    summary +=[f'{jl: <5}{sep} {indices[i]:>9d}{sep} \t' + \
                               f'{iso[i]:12.{PREC}f}{sep} ' +\
                               f'{aniso[i]:13.{PREC}f}{sep} {red_aniso[i]:17.{PREC}f}{sep}' + \
                               f'{asymm[i]:9.{PREC}f}{sep}' + \
                               f'{span[i]:8.{PREC}f}{sep}{skew[i]:5.{PREC}f}{sep}' + \
                               f'\t{a:9.{PREC}f}{sep} {b:9.{PREC}f}{sep} {c:9.{PREC}f}']
                    if args.quiet:
                        # add file path to end if in quiet mode
                        summary += [f]
                # output ms summary
                if args.csv:
                    fileout = 'MS_' + basename.replace('.magres', '.csv')
                    prefix = args.prefix
                    if prefix != "":
                        fileout = f"{prefix}_{fileout}"
                    fileout = os.path.join(dirname, fileout)
                    with open(fileout, 'w') as fd:
                        fd.write('\n'.join(summary))
                else:
                    print('\n'.join(summary))
            except KeyError:
                warnings.warn(f'No MS data found in {f}\n'
                'Set argument `-p efg` if the file(s) only contains EFG data ')
            except:
                warnings.warn('Failed to load MS data from .magres')
                raise

        # end ms

        # --- EFG --- #
        if 'efg' in properties:
            # list to hold the summary for this file
            summary = []

            try:

                vzz = EFGVzz.get(atoms)
                # For quadrupolar constants, isotopes become relevant. This means we need to create custom Property instances to
                # specify them. There are multiple ways to do so - check the docstrings for more details - but here we set them
                # by element. When nothing is specified it defaults to the most common NMR active isotope.

                isotopes = args.isotope
                qP = EFGQuadrupolarConstant(isotopes=isotopes) # Deuterated; for the others use the default
                qC = qP(atoms)/1e6 # To MHz

                quat = EFGQuaternion.get(atoms)
                eta  = EFGAsymmetry.get(atoms)

                # if not quiet, print header
                if not args.quiet:
                    if not args.csv:
                        summary += [EFG_HEADER.format(fname=f)]
                    table_headers = ['Label', 'Index', 'Vzz/au', 'Cq/MHz', 'Eta', 'alpha/deg', 'beta/deg', 'gamma/deg']
                    summary += [f'{sep}'.join([f'{lab:^9}' for lab in table_headers])]

                for i, jl in enumerate(labels):
                    a, b, c = quat[i].euler_angles(mode=args.euler)*180/np.pi # rad to degrees
                    summary +=[f'{jl:<5}{sep} {indices[i]:>9d}{sep}\t' +\
                               f'{vzz[i]:9.{PREC}f}{sep} {qC[i]:12.{PREC}e}{sep}' + \
                               f'{eta[i]:8.{PREC}f}{sep}' + \
                               f'{a:9.{PREC}f}{sep} {b:9.{PREC}f}{sep} {c:9.{PREC}f}']
                    if args.quiet:
                        # add file path to end if in quiet mode
                        summary[-1] += f'  {f}'
                # output EFG summary
                if args.csv:
                    fileout = 'EFG_' + basename.replace('.magres', '.csv')
                    prefix = args.prefix
                    if prefix != "":
                        fileout = f"{prefix}_{fileout}"
                    fileout = os.path.join(dirname, fileout)
                    with open(fileout, 'w') as fd:
                        fd.write('\n'.join(summary))
                else:
                    print('\n'.join(summary))
            except KeyError:
                warnings.warn(f'No EFG data found in {f}\n'
                'Set argument `-p ms` if the file(s) only contains MS data ')
            except:
                warnings.warn(f'Failed to load EFG data from .magres, {f}')
                raise

        # end EFG section
        ###############




if __name__ == '__main__':
    sys.exit(__main__())



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of AtomsProperties that relate to NMR dipole-dipole
couplings"""


import numpy as np

from soprano.data.nmr import _get_isotope_data
from soprano.nmr import NMRTensor
from soprano.nmr.utils import _dip_constant, _dip_tensor
from soprano.properties import AtomsProperty
from soprano.rnd import Random
from soprano.selection import AtomSelection
from soprano.utils import minimum_periodic, minimum_supcell, supcell_gridgen


class DipolarCoupling(AtomsProperty):

    """
    DipolarCoupling

    Produces a dictionary of dipole-dipole coupling constants for atom pairs
    in the system. For each pair, the closest periodic copy will be considered.
    The constant for a pair of nuclei i and j is defined as:

    .. math::

        d_{ij} = -\\frac{\\mu_0\\hbar\\gamma_i\\gamma_j}{8\\pi^2r_{ij}^3}

    where the gammas represent the gyromagnetic ratios of the nuclei and the
    r is their distance. The full tensor of the interaction is then defined as

    .. math::

         D_{ij} =
         \\begin{bmatrix}
          -d_{ij}   & 0         & 0       \\\\
          0         & -d_{ij}   & 0       \\\\
          0         & 0         & 2d_{ij} \\\\
         \\end{bmatrix}

    where the z-axis is aligned with :math:`r_{ij}` and the other two can be any
    directions in the orthogonal plane.

    Parameters:
      sel_i (AtomSelection or [int]): Selection or list of indices of atoms
                                      for which to compute the dipolar
                                      coupling. By default is None
                                      (= all of them).
      sel_j (AtomSelection or [int]): Selection or list of indices of atoms
                                      for which to compute the dipolar
                                      coupling with the ones in sel_i. By
                                      default is None (= same as sel_i).
      isotopes (dict): dictionary of specific isotopes to use, by element
                       symbol. If the isotope doesn't exist an error will
                       be raised.
      isotope_list (list): list of isotopes, atom-by-atom. To be used if
                           different atoms of the same element are supposed
                           to be of different isotopes. Where a 'None' is
                           present will fall back on the previous
                           definitions. Where an isotope is present it
                           overrides everything else.
      self_coupling (bool): if True, include coupling of a nucleus with its
                            own closest periodic copy. Otherwise excluded.
                            Default is False.
      block_size (int): maximum size of blocks used when processing large
                        chunks of pairs. Necessary to avoid memory problems
                        for very large systems. Default is 1000.
      isonuclear (bool): if True, only compute couplings between nuclei of
                         the same element. Default is False.

    Returns:
      dip_dict (dict): Dictionary of couplings in Hz and r_{ij} versors,
                       pointing from i to j, by atomic index pair.

    """

    default_name = "dip_coupling"
    default_params = {
        "sel_i": None,
        "sel_j": None,
        "isotopes": {},
        "isotope_list": None,
        "self_coupling": False,
        "block_size": 1000,
        "isonuclear": False,
    }

    @staticmethod
    def extract(
        s, sel_i, sel_j, isotopes, isotope_list, self_coupling, block_size, isonuclear
    ):
        # Selections
        if sel_i is None:
            sel_i = AtomSelection.all(s)
        elif not isinstance(sel_i, AtomSelection):
            sel_i = AtomSelection(s, sel_i)

        if sel_j is None:
            sel_j = sel_i
        elif not isinstance(sel_j, AtomSelection):
            sel_j = AtomSelection(s, sel_j)

        # Find gammas
        elems = s.get_chemical_symbols()

        gammas = _get_isotope_data(elems, "gamma", isotopes, isotope_list)

        # Viable pairs
        pairs = [(i, j) for i in sel_i.indices for j in sel_j.indices]
        if not self_coupling:
            pairs = [p for p in pairs if p[0] != p[1]]

        if isonuclear:
            pairs = [p for p in pairs if elems[p[0]] == elems[p[1]]]

        pairs = np.array(pairs).T
        # Need to sort them and remove any duplicates, also take i < j as
        # convention
        pairs = np.array(list(zip(*set([tuple(x) for x in np.sort(pairs, axis=0).T]))))

        pos = s.get_positions()

        # Split this in blocks to make sure we don't clog the memory

        d_ij = np.zeros((0,))
        v_ij = np.zeros((0, 3))

        # check if empty
        if pairs.shape[0] == 0:
            return {}

        npairs = pairs.shape[1]

        for b_i in range(0, npairs, block_size):
            block = pairs.T[b_i : b_i + block_size]
            r_ij = pos[block[:, 1]] - pos[block[:, 0]]
            # Reduce to NN
            r_ij, _ = minimum_periodic(r_ij, s.get_cell(), exclude_self=True)
            # Distance
            R_ij = np.linalg.norm(r_ij, axis=1)
            # Versors
            v_ij = np.concatenate([v_ij, r_ij / R_ij[:, None]], axis=0)
            # Couplings
            d_ij = np.concatenate(
                [
                    d_ij,
                    _dip_constant(
                        R_ij * 1e-10, gammas[block[:, 0]], gammas[block[:, 1]]
                    ),
                ]
            )

        return {tuple(ij): [d_ij[l], v_ij[l]] for l, ij in enumerate(pairs.T)}


class DipolarTensor(AtomsProperty):

    """
    DipolarTensor

    Produces a dictionary of dipole-dipole coupling tensors for atom pairs
    in the system. For each pair, the closest periodic copy will be considered.
    The coupling constant for a pair of nuclei i and j is defined as:

    .. math::

        d_{ij} = -\\frac{\\mu_0\\hbar\\gamma_i\\gamma_j}{8\\pi^2r_{ij}^3}

    where the gammas represent the gyromagnetic ratios of the nuclei and the
    r is their distance. The full tensor of the interaction is then defined as

    .. math::

         D_{ij} = d_{ij}(3\\hat{r}_{ij}\\otimes \\hat{r}_{ij}-\\mathbb{I})

    where :math:`\\hat{r}_{ij} = r_{ij}/|r_{ij}|` and the Kronecker product is
    used.

    Parameters:
      sel_i (AtomSelection or [int]): Selection or list of indices of atoms
                                      for which to compute the dipolar
                                      coupling. By default is None
                                      (= all of them).
      sel_j (AtomSelection or [int]): Selection or list of indices of atoms
                                      for which to compute the dipolar
                                      coupling with the ones in sel_i. By
                                      default is None (= same as sel_i).
      isotopes (dict): dictionary of specific isotopes to use, by element
                       symbol. If the isotope doesn't exist an error will
                       be raised.
      isotope_list (list): list of isotopes, atom-by-atom. To be used if
                           different atoms of the same element are supposed
                           to be of different isotopes. Where a 'None' is
                           present will fall back on the previous
                           definitions. Where an isotope is present it
                           overrides everything else.
      self_coupling (bool): if True, include coupling of a nucleus with its
                            own closest periodic copy. Otherwise excluded.
                            Default is False.
      block_size (int): maximum size of blocks used when processing large
                        chunks of pairs. Necessary to avoid memory problems
                        for very large systems. Default is 1000.
      rotation_axis (np.ndarray): if present, return the residual dipolar
                                  tensors after fast averaging around the
                                  given axis. Default is None.
      isonuclear (bool): if True, only compute couplings between nuclei of
                         the same element. Default is False.

    Returns:
      dip_dict (dict): Dictionary of `NMRTensor` objects in Hz by atomic index pair.

    """

    default_name = "dip_coupling"
    default_params = {
        "sel_i": None,
        "sel_j": None,
        "isotopes": {},
        "isotope_list": None,
        "self_coupling": False,
        "block_size": 1000,
        "rotation_axis": None,
        "isonuclear": False,
    }

    @staticmethod
    def extract(
        s,
        sel_i,
        sel_j,
        isotopes,
        isotope_list,
        self_coupling,
        block_size,
        rotation_axis,
        isonuclear,
    ):
        dip_dict = DipolarCoupling.extract(
            s,
            sel_i=sel_i,
            sel_j=sel_j,
            isotopes=isotopes,
            isotope_list=isotope_list,
            self_coupling=self_coupling,
            block_size=block_size,
            isonuclear=isonuclear,
        )

        # Now build the tensors
        tdict = {}
        for ij, (d, r) in dip_dict.items():
            tdict[ij] = NMRTensor(_dip_tensor(d, r, rotation_axis), order='n')

        return tdict


class DipolarDiagonal(AtomsProperty):

    """
    DipolarDiagonal

    Produces a dictionary of dipole-dipole tensors as eigenvalues and
    eigenvectors for atom pairs in the system. For each pair, the closest
    periodic copy will be considered.

    Parameters:
      sel_i (AtomSelection or [int]): Selection or list of indices of atoms
                                      for which to compute the dipolar
                                      coupling. By default is None
                                      (= all of them).
      sel_j (AtomSelection or [int]): Selection or list of indices of atoms
                                      for which to compute the dipolar
                                      coupling with the ones i sel_i. By
                                      default is None (= same as sel_i).
      isotopes (dict): dictionary of specific isotopes to use, by element
                       symbol. If the isotope doesn't exist an error will
                       be raised.
      isotope_list (list): list of isotopes, atom-by-atom. To be used if
                           different atoms of the same element are supposed
                           to be of different isotopes. Where a 'None' is
                           present will fall back on the previous
                           definitions. Where an isotope is present it
                           overrides everything else.
      self_coupling (bool): if True, include coupling of a nucleus with its
                            own closest periodic copy. Otherwise excluded.
                            Default is False.
      block_size (int): maximum size of blocks used when processing large
                        chunks of pairs. Necessary to avoid memory problems
                        for very large systems. Default is 1000.
      isonuclear (bool): if True, only compute couplings between nuclei of
                         the same element. Default is False.

    Returns:
      dip_tens_dict (dict): Dictionary of dipolar eigenvalues (in Hz) and
                            eigenvectors, by atomic index pair.

    """

    default_name = "dip_diagonal"
    default_params = {
        "sel_i": None,
        "sel_j": None,
        "isotopes": {},
        "isotope_list": None,
        "self_coupling": False,
        "block_size": 1000,
        "isonuclear": False,
    }

    @staticmethod
    def extract(
        s, sel_i, sel_j, isotopes, isotope_list, self_coupling, block_size, isonuclear
    ):
        # First, just get the values
        dip_dict = DipolarCoupling.extract(
            s,
            sel_i,
            sel_j,
            isotopes,
            isotope_list,
            self_coupling,
            block_size,
            isonuclear,
        )

        # Now build the tensors
        dip_tens_dict = {}

        for ij, (d, v) in dip_dict.items():
            evals = np.array([-d, -d, 2 * d])
            # Eigenvectors
            evecs = np.zeros((3, 3))
            # Z is equal to v
            evecs[:, 2] = v
            # Y is any random orthogonal vector
            rv = Random.random(3)
            evecs[:, 1] = np.cross(v, rv)
            evecs[:, 1] /= np.linalg.norm(evecs[:, 1])
            # X = Y cross Z
            evecs[:, 0] = np.cross(evecs[:, 1], v)

            dip_tens_dict[ij] = {"evals": evals, "evecs": evecs}

        return dip_tens_dict


class DipolarRSS(AtomsProperty):

    """
    DipolarRSS

    Compute the Dipolar constant Root Sum Square for each atom in a system,
    including periodicity, within a cutoff.

    Parameters:
      cutoff (float): cutoff radius in Angstroms at which the sum stops. By
                      default 5 Ang.
      isonuclear (bool): if True, only nuclei of the same species will be
                         considered. By default is False.
      isotopes (dict): dictionary of specific isotopes to use, by element
                       symbol. If the isotope doesn't exist an error will
                       be raised.
      isotope_list (list): list of isotopes, atom-by-atom. To be used if
                           different atoms of the same element are supposed
                           to be of different isotopes. Where a 'None' is
                           present will fall back on the previous
                           definitions. Where an isotope is present it
                           overrides everything else.

    Returns:
      dip_rss (np.ndarray): dipolar constant RSS for each atom in the system

    """

    default_name = "dip_rss"
    default_params = {
        "cutoff": 5.0,
        "isonuclear": False,
        "isotopes": {},
        "isotope_list": None,
    }

    @staticmethod
    def extract(s, cutoff, isonuclear, isotopes, isotope_list):
        # Supercell size
        scell_shape = minimum_supcell(cutoff, s.get_cell())
        _, scell = supcell_gridgen(s.get_cell(), scell_shape)

        pos = s.get_positions()
        elems = np.array(s.get_chemical_symbols())

        gammas = _get_isotope_data(elems, "gamma", isotopes, isotope_list)

        dip_rss = []

        for i, el in enumerate(elems):
            # Distances?
            if not isonuclear:
                rij = pos.copy()
                gj = np.tile(gammas, len(scell))
            else:
                rij = pos[np.where(elems == el)]
                gj = gammas[i]
            rij = rij[None, :, :] + scell[:, None, :] - pos[i, None, None]
            Rij = np.linalg.norm(rij.reshape((-1, 3)), axis=-1)
            # Valid indices?
            ij = np.where((Rij > 0) & (Rij <= cutoff))
            Rij = Rij[ij] * 1e-10
            try:
                gj = gj[ij]
            except IndexError:
                pass

            dip = _dip_constant(Rij, gammas[i], gj)
            dip_rss.append(np.sqrt(np.sum(dip**2)))

        return np.array(dip_rss)



class DipolarEuler(AtomsProperty):

    """
    DipolarEuler

    Produces a dictionary of dipole-dipole coupling tensor Euler angles for atom pairs
    in the system. See the DipolarTensor property for details on the tensor.

    Parameters:
        sel_i (AtomSelection or [int]): Selection or list of indices of atoms
                                        for which to compute the dipolar
                                        coupling. By default is None
                                        (= all of them).
        sel_j (AtomSelection or [int]): Selection or list of indices of atoms
                                        for which to compute the dipolar
                                        coupling with the ones in sel_i. By
                                        default is None (= same as sel_i).
        isotopes (dict): dictionary of specific isotopes to use, by element
                        symbol. If the isotope doesn't exist an error will
                        be raised.
        isotope_list (list): list of isotopes, atom-by-atom. To be used if
                            different atoms of the same element are supposed
                            to be of different isotopes. Where a 'None' is
                            present will fall back on the previous
                            definitions. Where an isotope is present it
                            overrides everything else.
        self_coupling (bool): if True, include coupling of a nucleus with its
                                own closest periodic copy. Otherwise excluded.
                                Default is False.
        isonuclear (bool): if True, only compute couplings between nuclei of
                            the same element. Default is False.
        block_size (int): maximum size of blocks used when processing large
                            chunks of pairs. Necessary to avoid memory problems
                            for very large systems. Default is 1000.
        rotation_axis (np.ndarray): if present, return the residual dipolar
                                    tensors after fast averaging around the
                                    given axis. Default is None.
        convention (str): 'zyz' or 'zxz' accepted - the ordering of the Euler
                        angle rotation axes. Default is zyz 
        passive (bool):  active or passive rotations. Default is active (passive=False)
        degrees (bool):  return the angles in degrees. Default is False

    Returns:
        dip_dict (dict): Dictionary of tensors in Hz by atomic index pair.

    """

    default_name = "dip_eulers"
    default_params = {
        "sel_i": None,
        "sel_j": None,
        "isotopes": {},
        "isotope_list": None,
        "self_coupling": False,
        "isonuclear": False,
        "block_size": 1000,
        "rotation_axis": None,
        "convention": "zyz",
        "passive": False,
        "degrees": False
    }

    @staticmethod
    def extract(
        s,
        sel_i,
        sel_j,
        isotopes,
        isotope_list,
        self_coupling,
        isonuclear,
        block_size,
        rotation_axis,
        convention,
        passive,
        degrees
    ):

        dip_dict = DipolarTensor.extract(
            s,
            sel_i=sel_i,
            sel_j=sel_j,
            isotopes=isotopes,
            isotope_list=isotope_list,
            self_coupling=self_coupling,
            isonuclear=isonuclear,
            block_size=block_size,
            rotation_axis=rotation_axis,
        )

        # Now build the euler angles dict
        eulerdict = {}
        for ij, T in dip_dict.items():
            eulers = T.euler_angles(convention=convention, passive=passive, degrees=degrees)
            eulerdict[ij] = eulers

        return eulerdict



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""CLI to extract and process NMR-related properties from .magres files.

TODO: add support for different shift {Haeberlen,NQR,IUPAC}and quadrupole {Haeberlen,NQR} conventions.
TODO: check if df is too wide to fit in window -- if so, split into multiple plots.
TODO: spinsys output is not yet implemented.
TODO: document config file setup

REFACTORED NMR_EXTRACRT ETC., NEED TO MAKE SURE THE OTHER CLI COMMANDS STILL WORK 
"""

__author__ = "J. Kane Shenton"
__maintainer__ = "J. Kane Shenton"
__email__ = "kane.shenton@stfc.ac.uk"
__date__ = "July 08, 2022"


import logging
from typing import List, Optional

import click
import click_log
import numpy as np
import pandas as pd
from ase import Atoms, io
from ase.units import Bohr, Ha

from soprano.data.nmr import _get_isotope_list
from soprano.properties.labeling import MagresViewLabels, UniqueSites
from soprano.properties.nmr import *
from soprano.scripts.cli_utils import (
    NMREXTRACT_OPTIONS,
    NO_CIF_LABEL_WARNING,
    add_options,
    apply_df_filtering,
    average_quaternions_by_tags,
    expand_aliases,
    find_XHn_groups,
    print_results,
    reload_as_molecular_crystal,
    sortdf,
    units_rename,
    viewimages,
)
from soprano.selection import AtomSelection
from soprano.utils import has_cif_labels, merge_sites

# logging
logging.captureWarnings(True)
logger = logging.getLogger("cli")
click_log.basic_config(logger)

HEADER = """
##########################################
#  Extracting NMR info from magres file  #
"""
FOOTER = """
# End of NMR info extraction            #
##########################################
"""

MS_MINIMAL_COLUMNS = ["MS_shielding", "MS_anisotropy"]
EFG_MINIMAL_COLUMNS = ["EFG_quadrupolar_constant", "EFG_asymmetry",]
NMR_COLUMN_ALIASES = {
    "minimal" : MS_MINIMAL_COLUMNS + EFG_MINIMAL_COLUMNS,
    "ms": MS_MINIMAL_COLUMNS,
    "efg": EFG_MINIMAL_COLUMNS,
    "angles": [
        "alpha", "beta", "gamma"
    ],
    "essential": ["labels", "species", "multiplicity", "tags", "file"]
}

@click.command()
# one of more files
@click.argument("files", nargs=-1, type=click.Path(exists=True), required=True)
@add_options(NMREXTRACT_OPTIONS)
def nmr(
    files,
    subset=None,
    output=None,
    output_format=None,
    merge=False,
    isotopes={},
    references={},
    gradients={},
    reduce=True,
    average_group=None,
    symprec=1e-4,
    properties=["efg", "ms"],
    precision=3,
    euler_convention="zyz",
    sortby="",
    sort_order="ascending",
    include=None,
    exclude=None,
    query=None,
    view=False,
    verbosity=0,
    ms_tag="ms",
    efg_tag="efg",
):
    """
    Extract and analyse NMR data from magres file(s) or extended XYZ files.

    Usage:
    soprano nmr seedname.magres
    soprano nmr data.xyz --ms-tag pred_ms --efg-tag ref_efg

    Processes .magres file(s) or extended XYZ files containing NMR-related properties
    and prints a summary. It defaults to printing all NMR properties
    present in the file for all the atoms.

    See the below arguments for how to extract specific information.
    """
    if verbosity == 0:
        logging.basicConfig(level=logging.WARNING)
    elif verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.DEBUG)

    dfs, images = nmr_extract_multi(
        files,
        subset=subset,
        merge=merge,
        isotopes=isotopes,
        references=references,
        gradients=gradients,
        reduce=reduce,
        average_group=average_group,
        symprec=symprec,
        properties=properties,
        euler_convention=euler_convention,
        sortby=sortby,
        sort_order=sort_order,
        include=include,
        exclude=exclude,
        query=query,
        logger=logger,
        ms_tag=ms_tag,
        efg_tag=efg_tag,
    )
    if view:
        viewimages(images)

    # write to file(s)
    print_results(dfs, output, output_format, precision, verbosity > 0)


def nmr_extract_multi(
    files,
    merge=False,
    logger=None,
    sortby=None,
    sort_order="ascending",
    isotopes={},
    references={},
    gradients={},
    properties=["efg", "ms"],
    euler_convention="zyz",
    include=[],
    exclude=[],
    query="",
    **kwargs,
):
    """
    Extract NMR data from magres file(s). See CLI help for more details on the arguments. (`soprano nmr --help`)

    Args:
        files (list): list of magres files to extract data from.
        merge (bool): whether to merge the pandas dataframes from more than one .magres file.
        logger (logging.Logger): logger to use for logging. If None, a new logger is created.
        sortby (str): column to sort the dataframe by.
        sort_order (str): order to sort the dataframe by. Options are 'ascending' or 'descending'.
        isotopes (dict): dictionary of isotope labels to use for each element. e.g. {"H": "1H", "C": "13C"}.
        references (dict): dictionary of shielding reference values to use for each element. e.g. {"H": 30.0, "C": 100.0}.
        gradients (dict): dictionary of gradient reference values to use for each element. e.g. {"H": -1.0, "C": -0.95}.
                          If not provided, the gradient is assumed to be -1 for all elements.
        properties (list): list of properties to extract. Options are 'efg', 'ms'.
        euler_convention (str): convention to use for Euler angles. Options are 'zyz' or 'zxz'.
        include (str): comma-separated list of columns to include in the output.
        exclude (str): comma-separated list of columns to exclude from the output.
        query (str): query string to filter the dataframe.
    Expected kwargs:
        subset (str): subset of atoms to extract data from. e.g. "H1,H2,C" for all H1, H2 and C atoms.
        reduce (bool): whether to reduce to symmetry equivalent sites (using either the CIF labels or symmetry operations found using SPGLIB).
        average_group (str): comma-separated list of functional groups to average over e.g. methyl groups. e.g. "CH3" averages over H atoms in methyl groups.
        merging_strategies (dict): dictionary of merging strategies to use for each property. e.g. {"positions": lambda x: x[0]}.
        symprec (float): tolerance for symmetry operations. Default is 1e-4.



    Returns:
        dfs (list): list of pandas DataFrames containing the extracted data.
        images (list): list of ASE Atoms objects containing the crystal structures.
    """

    if logger is None:
        logger = logging.getLogger(__name__)

    if isotopes:
        logger.info(f"\nUsing custom isotopes for: {isotopes}")

    dfs = []
    images = []
    # loop over files
    for fname in files:
        logger.info(HEADER)
        logger.info(fname)
        logger.info(f"\nExtracting properties: {properties}")

        # try to read in the file:
        try:
            atoms = io.read(fname)
            # immediately try to reload as molecular crystal
            atoms = reload_as_molecular_crystal(atoms)
            # label atoms
            atoms = label_atoms(atoms)
        except OSError:
            logger.error(f"Could not read file {fname}, skipping.")
            return

        # Create mapping from properties to their corresponding tags
        property_tags = {
            'ms': kwargs.get('ms_tag', 'ms'),
            'efg': kwargs.get('efg_tag', 'efg')
        }
        
        # Do they actually have any magres data?
        required_tags = [property_tags[prop] for prop in properties if prop in property_tags]
        if not any([atoms.has(tag) for tag in required_tags]):
            logger.error(
                f"File {fname} has no {' or '.join(required_tags)} data to extract. Skipping."
            )
            continue

        atoms = nmr_extract_atoms(
            atoms,
            logger=logger,
            **kwargs,
        )

        if atoms is None:
            # This can happen for example if the given magres file has no
            # atoms that match the selection string
            continue

        # build the dataframe
        df = build_nmr_df(
            atoms,
            fname,
            isotopes=isotopes,
            references=references,
            gradients=gradients,
            properties=properties,
            euler_convention=euler_convention,
            property_tags=property_tags,
            logger=logger,
        )

        if len(properties) == 1 and include is not None:
            # we're looking for only one property
            # so we need the relevant subset of minimal
            if "minimal" in include:
                # replace it with properties[0]
                include.pop(include.index("minimal"))
                include += properties

        # apply filters
        df = apply_df_filtering(
            df,
            expand_aliases(include, NMR_COLUMN_ALIASES),
            exclude,
            query,
            essential_columns=NMR_COLUMN_ALIASES["essential"],
            logger=logger,
        )

        # If length of df is 0, we skip this file
        if len(df) == 0:
            logger.warning(
                f"No results found for {fname}.\n "
                "Try removing filters/checking the file contents."
            )
            continue

        # ----- atoms object manipulation -----
        # only keep the atoms that are in the dataframe (based on tag)
        atoms = atoms[np.isin(atoms.get_tags(), np.array(df["tags"].values))]

        # if the df is not empty, append it to the list
        if len(df) > 0:
            dfs.append(df)
            images.append(atoms)
            logger.info(FOOTER)
        # if the df is empty, raise warning and don't append
        else:
            logger.warning(
                f"No results found for {fname}.\n "
                "Try removing filters/checking the file contents."
            )

    if merge:
        # merge all dataframes into one
        dfs = [pd.concat(dfs, axis=0)]
    for i, df in enumerate(dfs):
        dfs[i] = sortdf(df, sortby, sort_order)
    # rename columns to include units for those that have units
    for df in dfs:
        df.rename(columns=units_rename, inplace=True)
    return dfs, images


def nmr_extract_atoms(
    atoms: Atoms,
    subset: str = "",
    reduce=True,
    average_group: str = "",
    merging_strategies: dict = {
        # for the positions, just take the first one
        "positions": lambda x: x[0],
        # for the labels, just take the first one
        "labels": lambda x: x[0],
    },
    symprec: float = 1e-4,
    ms_tag: str = "ms",
    efg_tag: str = "efg",
    logger: logging.Logger = logging.getLogger("cli"),
):
    """
    Extract NMR data from a single ASE Atoms object.

    Args:
        atoms (Atoms): the Atoms object to extract data from.
        subset (str): subset of atoms to extract data from. e.g. "H1,H2,C" for all H1, H2 and C atoms.
        reduce (bool): whether to reduce to symmetry equivalent sites (using either the CIF labels or symmetry operations found using SPGLIB).
        average_group (str): comma-separated list of functional groups to average over e.g. methyl groups. e.g. "CH3" averages over H atoms in methyl groups.
        merging_strategies (dict): dictionary of merging strategies to use for each property. e.g. {"positions": lambda x: x[0]}.
        symprec (float): tolerance for symmetry operations. Default is 1e-4.
        logger (logging.Logger): logger to use for logging. If not provided, we use the default logger for the cli.

    Returns:
        atoms (Atoms): the (subset of the) Atoms object with the extracted data.
    """


    # create new array for multiplicity
    multiplicity = np.ones(len(atoms), dtype=int)
    atoms.set_array("multiplicity", multiplicity)

    # reduce by symmetry?
    tags = np.arange(len(atoms))

    if reduce:
        logger.info("\nTagging equivalent sites")
        # tag equivalent sites
        tags = UniqueSites.get(atoms, symprec=symprec)

        # log the number of unique sites
        unique_sites, unique_site_idx = np.unique(tags, return_index=True)
        logger.debug(f"    This leaves {len(unique_sites)} unique sites")
        if atoms.has("labels"):
            labels = np.asarray(atoms.get_array("labels"), dtype="U25")
            logger.debug(f"    The unique site labels are: {labels[unique_site_idx]}")

    # check to make sure that all sites with the same tag have the same MSIsotropy
    # if not, throw a warning, suggest to turn on debug logging and --no-reduce flag
    # and then continue
    if atoms.has(ms_tag) and not check_equivalent_sites_ms(atoms, tags, tag=ms_tag):
        logger.warning(
            "    Some sites with the same symmetry tag/CIF label have different MS isotropy values."
        )
        logger.warning(
            "    You can turn off symmetry reduction with the --no-reduce flag."
        )
        logger.warning("    You can also turn on debug logging with the -vv flag.")
        logger.warning(
            "    If you find that the (symmetry) reduction algorithm is working incorrectly,"
        )
        logger.warning("    please report this to the developers.")

    # set tags to atoms object
    atoms.set_tags(tags)

    if average_group:
        atoms = tag_functional_groups(average_group, atoms, vdw_scale=1.0)


    all_selections = AtomSelection.all(atoms)
    # select subset of atoms based on selection string
    if subset:
        logger.info(f"\nSelecting atoms based on selection subset string: {subset}")
        try:
            sel_selectionstring = AtomSelection.from_selection_string(atoms, subset)
            all_selections *= sel_selectionstring
        except ValueError as e:
            logger.error(f"Could not select atoms based on selection string: {e}")
            return

        logger.debug(f"    Selected atoms: {all_selections.indices}")
        ## apply selection string to atoms object
        atoms = all_selections.subset(atoms)

    atoms = merge_tagged_sites(atoms, merging_strategies=merging_strategies)

    return atoms


def label_atoms(atoms: Atoms) -> Atoms:
    if has_cif_labels(atoms):
        return atoms

    # Inform user of best practice RE CIF labels
    logger.debug(NO_CIF_LABEL_WARNING)

    if atoms.has("magresview_labels"):
        labels = atoms.get_array("magresview_labels")
    else:
        labels = MagresViewLabels.get(atoms, store_array=True)

    # note we must change datatype to allow more space!
    labels = np.array(labels, dtype="U25")

    # remove current labels:
    if atoms.has("labels"):
        atoms.set_array("labels", None)
    # add labels to atoms object
    atoms.set_array("labels", labels)
    return atoms


def tag_functional_groups(
    average_group: str,
    atoms: Atoms,
    vdw_scale: float = 1.0,
) -> Atoms:
    """
    Average over groups of atoms based on the average_group string.
    See find_XHn_groups for more details.

    Args:
        average_group (str): string of comma-separated patterns to average over. e.g. 'CH3,CH2'
        atoms (Atoms): Atoms object
        vdw_scale (float): scaling factor for the van der Waals radii. Default is 1.0.

    Returns:
        Atoms
    """
    if atoms.has("tags"):
        tags = atoms.get_tags()
    else:
        tags = np.arange(len(atoms))

    labels = atoms.get_array("labels")
    # make sure dtype of labels allows for enough characters
    labels = labels.astype("U25")
    XHn_groups = find_XHn_groups(atoms, average_group, tags=tags, vdw_scale=vdw_scale)
    logger.info(f"\nAveraging over functional groups: {average_group}")
    for ipat, pattern in enumerate(XHn_groups):
        # check if we found any that matched this pattern
        if len(pattern) == 0:
            logging.warning(
                f"No XHn groups found for pattern {average_group.split(',')[ipat]}"
            )
            continue
        logger.debug(f"Found {len(pattern)} {average_group.split(',')[ipat]} groups")
        # get the indices of the atoms that matched this pattern
        # update the tags and labels accordingly
        for ig, group in enumerate(pattern):
            logger.debug(f"    Group {ig} contains: {np.unique(labels[group])}")
            # fix labels here as aggregate of those in group
            combined_label = ",".join(np.unique(labels[group]))
            # labels[group] = f'{ig}'#combined_label
            labels[group] = combined_label

            tags[group] = -(ipat + 1) * 1e5 - ig
    # update atoms object with new labels
    # note we must change datatype to allow more space!
    atoms.set_array("labels", None)
    atoms.set_array("labels", labels, dtype="U25")
    # update atoms tags
    atoms.set_tags(tags)
    return atoms


def merge_tagged_sites(atoms_in: Atoms, merging_strategies: dict = {}) -> Atoms:
    """
    Merge sites that are tagged with the same tag.

    Args:
        atoms (Atoms): Atoms object. Must have tags.
        merging_strategies (dict): dictionary of merging strategies. See merge_sites for more details.
    """
    atoms = atoms_in.copy()
    # if there are no tags present, return the atoms object
    if not atoms.has("tags"):
        return atoms

    # now we need to apply the filters etc to the atoms object
    # first we need to merge sites with the same tag
    unique_tags, unique_counts = np.unique(atoms.get_tags(), return_counts=True)
    # groups with more than one atom
    multi_group_tags = unique_tags[unique_counts > 1]
    for tag in multi_group_tags:
        # where are these tags in the current tags?
        tag_idx = np.where(atoms.get_tags() == tag)[0]
        # merge the sites
        atoms = merge_sites(
            atoms, tag_idx, merging_strategies=merging_strategies, keep_all=False
        )

    # sort by tag
    atoms = atoms[np.argsort(atoms.get_tags())]

    return atoms


def build_nmr_df(
    atoms: Atoms,
    fname: str,
    isotopes: dict = {},
    references: dict = {},
    gradients: dict = {},
    properties: List[str] = ["efg", "ms"],
    euler_convention: str = "zyz",
    property_tags: dict = {},
    logger: logging.Logger = logging.getLogger("cli"),
):
    """
    Build the dataframe containing the NMR properties.

    Args:
        atoms (ASE Atoms object): the atoms object to be reloaded.
        fname (str): the filename of the file being processed.
        all_selections (AtomSelection): the AtomSelection object containing all selections.
        isotopes (dict): dictionary of isotopes to use for each element. e.g. {'H': 2, 'C': 13}
        references (dict): dictionary of shielding references for each element. e.g. {'H': 20.0, 'C': 100.0}
        gradients (dict): dictionary of gradients for each element. e.g. {'H': -1.0, 'C': -0.95} defaults to {} == -1 for all elements.
        average_group (str): string of comma-separated patterns to average over. e.g. 'CH3,CH2'
        properties (list): list of properties to extract. e.g. ['efg', 'ms']
        euler_convention (str): the euler convention to use for the EFG tensor. Options are 'zyz' or 'zxz'

    Returns:
        df (pandas DataFrame): the dataframe containing the NMR properties.
    """
    elements = atoms.get_chemical_symbols()
    isotopelist = _get_isotope_list(elements, isotopes=isotopes, use_q_isotopes=False)
    species = [f"{iso}{el}" for el, iso in zip(elements, isotopelist)]
    labels = np.asarray(atoms.get_array("labels"), dtype="U25")
    tags = atoms.get_tags()

    df = pd.DataFrame(
        {
            "indices": atoms.get_array("indices"),
            "original_index": np.arange(len(atoms)),
            "labels": labels,
            "species": species,
            "multiplicity": atoms.get_array("multiplicity"),
            "tags": tags,
        }
    )

    # If there are MagresView-style labels, add them in
    if atoms.has("magresview_labels"):
        df.insert(2, "MagresView_labels", atoms.get_array("magresview_labels"))

    # Let's add a column for the file name -- useful to keep track of
    # which file the data came from if merging multiple files.
    df["file"] = fname
    if "ms" in properties:
        ms_tag = property_tags.get('ms', 'ms')
        try:
            ms_summary = pd.DataFrame(
                get_ms_summary(atoms, euler_convention, references, gradients, ms_tag)
            )
            if not references:
                # drop shift column if no references are given
                ms_summary.drop(columns=["MS_shift"], inplace=True)

            df = pd.concat([df, ms_summary], axis=1)
        except RuntimeError:
            logger.warning(
                f"No MS data found in {fname} with tag '{ms_tag}'\n"
                "Set argument `-p efg` if the file(s) only contains EFG data "
            )
        except:
            logger.warning("Failed to load MS data from .magres")
            raise
    if "efg" in properties:
        efg_tag = property_tags.get('efg', 'efg')
        try:
            efg_summary = pd.DataFrame(
                get_efg_summary(atoms, isotopes, euler_convention, efg_tag)
            )
            df = df = pd.concat([df, efg_summary], axis=1)
        except RuntimeError:
            logger.warning(
                f"No EFG data found in {fname} with tag '{efg_tag}'\n"
                "Set argument `-p ms` if the file(s) only contains MS data "
            )
        except:
            logger.warning("Failed to load EFG data from .magres")
            raise

    ## how many sites do we have now?
    total_explicit_sites = df["multiplicity"].sum()
    logger.info(f"\nFound {int(total_explicit_sites)} total sites.")
    logger.info(f"Reduced to {len(df)} sites.")

    return df


def get_ms_summary(
    atoms: Atoms,
    euler_convention: str,
    references: Optional[dict] = None,
    gradients: Optional[dict] = None,
    ms_tag: str = "ms",
) -> pd.DataFrame:
    """
    For an Atoms object with ms tensor arrays, return a summary of the tensors.

    Args:
        atoms (Atoms): the Atoms object
        euler_convention (str): the euler convention to use
        references (dict, optional): the reference tensors. Defaults to None. e.g. {'C': 100}
        gradients (dict, optional): the gradient tensors. Defaults to None. e.g. {'C': -1}
        ms_tag (str): the tag for the MS tensor array. Defaults to "ms".

    Returns:
        dict: a dictionary with the summary of the ms tensors
    """
    # Isotropy, Anisotropy and Asymmetry (Haeberlen convention)
    iso = MSIsotropy.get(atoms, tag=ms_tag)
    shift = MSIsotropy.get(atoms, ref=references, grad=gradients, tag=ms_tag)
    aniso = MSAnisotropy.get(atoms, tag=ms_tag)
    red_aniso = MSReducedAnisotropy.get(atoms, tag=ms_tag)
    asymm = MSAsymmetry.get(atoms, tag=ms_tag)
    # Span and skew
    span = MSSpan.get(atoms, tag=ms_tag)
    skew = MSSkew.get(atoms, tag=ms_tag)
    # quaternion
    quat = MSQuaternion.get(atoms, tag=ms_tag)
    # We need to be carefull with the angle averaging
    quat = average_quaternions_by_tags(quat, atoms.get_tags())
    # Euler angles
    alpha, beta, gamma = np.array(
        [q.euler_angles(mode=euler_convention) * 180 / np.pi for q in quat]
    ).T
    ms_summary = {
        "MS_shielding": iso,
        "MS_shift": shift,
        "MS_anisotropy": aniso,
        "MS_reduced_anisotropy": red_aniso,
        "MS_asymmetry": asymm,
        "MS_span": span,
        "MS_skew": skew,
        "MS_alpha": alpha,
        "MS_beta": beta,
        "MS_gamma": gamma,
    }
    return ms_summary


def get_efg_summary(
    atoms: Atoms,
    isotopes: dict,
    euler_convention: str,
    tag: str = "efg",
) -> dict:
    """
    For an Atoms object with EFG tensor arrays, return a summary of the tensors.

    Args:
        atoms (Atoms): the Atoms object
        isotopes (dict): the isotopes to use for the quadrupolar constants
        euler_convention (str): the euler convention to use
        tag (str): the tag for the EFG tensor array. Defaults to "efg".

    Returns:
        dict: a dictionary with the summary of the EFG tensors

    """
    Vzz = EFGVzz.get(atoms, tag=tag)
    # convert Vzz from au to V/m^2
    Vzz = Vzz * (Ha / Bohr) * 1e-1

    # For quadrupolar constants, isotopes become relevant. This means we need to create custom Property instances to
    # specify them. There are multiple ways to do so - check the docstrings for more details - but here we set them
    # by element. When nothing is specified it defaults to the most common NMR active isotope.
    qP = EFGQuadrupolarConstant(isotopes=isotopes, tag=tag)
    qC = qP(atoms) / 1e6  # To MHz

    # asymmetry
    eta = EFGAsymmetry.get(atoms, tag=tag)

    # quaternion
    quat = EFGQuaternion.get(atoms, tag=tag)
    # We need to be carefull with the angle averaging
    quat = average_quaternions_by_tags(quat, atoms.get_tags())
    # Euler angles
    alpha, beta, gamma = np.array(
        [q.euler_angles(mode=euler_convention) * 180 / np.pi for q in quat]
    ).T

    # NQR transitions
    nqrs = EFGNQR.get(atoms, isotopes=isotopes, tag=tag)
    # unique transitions
    transition_keys = sorted(set([k for nqr in nqrs for k in nqr.keys()]))
    nqr_dict = {}
    for k in transition_keys:
        header = f"EFG_NQR {k}"
        values = np.zeros(len(nqrs))
        for inqr, nqr in enumerate(nqrs):
            if k in nqr:
                values[inqr] = nqr[k] * 1e-6
            else:
                values[inqr] = np.nan
        nqr_dict[header] = values

    efg_summary = {
        "EFG_Vzz": Vzz,
        "EFG_quadrupolar_constant": qC,
        "EFG_asymmetry": eta,
        "EFG_alpha": alpha,
        "EFG_beta": beta,
        "EFG_gamma": gamma,
        **nqr_dict,
    }

    return efg_summary


def check_equivalent_sites_ms(atoms, tags, tolerance=1e-3, tag="ms"):
    """
    Check if the sites with the same tags have the same MS isotropy to within a tolerance.

    Args:
        atoms (Atoms): the Atoms object
        tags (list): the tags to check
        tolerance (float, optional): the tolerance. Defaults to 1e-3.
        tag (str): the tag for the MS tensor array. Defaults to "ms".

    Returns:
        bool: True if the sites are equivalent, False otherwise

    """
    unique_sites, counts = np.unique(tags, return_counts=True)
    ms = MSIsotropy.get(atoms, tag=tag)
    # loop over unique sites that have more than equivalent site
    for i in unique_sites[counts > 1]:
        # get the indices of the equivalent sites
        idx = np.where(tags == i)[0]
        # check if the ms isotropy is the same to within the tolerance
        if not np.allclose(ms[idx], ms[idx[0]], atol=tolerance):
            return False
    return True



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utility functions for NMR-related properties  
"""

import re
import warnings
from typing import List, Tuple, Union

import numpy as np
import scipy.constants as cnst
from ase.quaternions import Quaternion
from scipy.spatial.transform import Rotation

# Left here for backwards compatibility
from soprano.data.nmr import _get_isotope_data, _get_nmr_data, _el_iso

def _split_species(species: str) -> Tuple[int, str]:
        """
        Validate the species string and extract the isotope number and element.

        Args:
            species (str): The species string to validate. e.g. '13C' or '1H'.

        Returns:
            tuple: A tuple containing the isotope number (int), and element symbol (str, such as 'C' or 'H').
        """
        match = re.match(r"^(\d+)([A-Za-z]+)$", species)
        if not match:
            raise ValueError(f"Species must be an isotope symbol such as '13C' or '1H'. Got '{species}' instead.")
        isotope_number, element = match.groups()
        return int(isotope_number), element


def _evals_sort(evals, convention="c", return_indices=False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Sort a list of eigenvalue triplets by varios conventions"""
    evals = np.array(evals)
    iso = np.average(evals, axis=1)

    if convention in ("i", "d"):
        to_sort = evals
    elif convention == "h":
        to_sort = np.abs(evals - iso[:, None])
    elif convention == "n":
        # Although iso should be zero for NQR (EFG tensors are traceless),
        # we should really just sort by the absolute values of the eigenvalues

        # We can warn the user if the isotropic value is not zero
        if np.any(np.abs(iso) > 1e-10):
            warnings.warn("Isotropic value(s) are not zero but NQR order is requested.\n"
                "If you're dealing with an EFG tensor, "
                "then check it carefully since these should be traceless.\n"
                "Sorting by absolute values.\n"
                )
        to_sort = np.abs(evals)
    else:
        raise ValueError("Invalid convention. Must be one of 'i', 'd', 'h', or 'n'. Received: ", convention)

    sort_i = np.argsort(to_sort, axis=1)
    if convention == "d":
        sort_i = sort_i[:, ::-1]
    elif convention == "h":
        sort_i[:, 0], sort_i[:, 1] = sort_i[:, 1], sort_i[:, 0].copy()
    sorted_evals = evals[np.arange(evals.shape[0])[:, None], sort_i]
    if not return_indices:
        return sorted_evals
    else:
        return sorted_evals, sort_i


def _haeb_sort(evals, return_indices=False):
    return _evals_sort(evals, "h", return_indices)


def _anisotropy(haeb_evals, reduced=False):
    """Calculate anisotropy given eigenvalues sorted with Haeberlen
    convention"""

    f = 2.0 / 3.0 if reduced else 1.0

    return (haeb_evals[:, 2] - (haeb_evals[:, 0] + haeb_evals[:, 1]) / 2.0) * f


def _asymmetry(haeb_evals):
    """Calculate asymmetry"""

    aniso = _anisotropy(haeb_evals, reduced=True)
    # Fix the anisotropy zero values
    aniso = np.where(aniso == 0, np.inf, aniso)

    return (haeb_evals[:, 1] - haeb_evals[:, 0]) / aniso


def _span(evals):
    """Calculate span
    
    .. math::
        \\Omega = \\sigma_{33} - \\sigma_{11}

    where :math:`\\sigma_{33}` is the largest, and :math:`\\sigma_{11}` is the
    smallest eigenvalue.
    
    """

    return np.amax(evals, axis=-1) - np.amin(evals, axis=-1)


def _skew(evals):
    """Calculate skew

    .. math::
        \\kappa = 3 (\\sigma_{iso} - \\sigma_{22}) / \\Omega

    where :math:`\\Omega` is the span of the tensor.

    Note that for chemical shift tensors (:math:`\\delta`), the sign is reversed.
    """

    span = _span(evals)
    span = np.where(span == 0, np.inf, span)
    return 3 * (np.average(evals, axis=1) - np.median(evals, axis=1)) / span


def _evecs_2_quat(evecs):
    """Convert a set of eigenvectors to a Quaternion expressing the
    rotation of the tensor's PAS with respect to the Cartesian axes"""

    # First, guarantee that the eigenvectors express *proper* rotations
    evecs = np.array(evecs) * np.linalg.det(evecs)[:, None, None]

    # Then get the quaternions
    return [Quaternion.from_matrix(evs.T) for evs in evecs]


def _dip_constant(Rij, gi, gj):
    """Dipolar constants for pairs ij, with distances Rij and gyromagnetic
    ratios gi and gj"""

    return -(cnst.mu_0 * cnst.hbar * gi * gj / (8 * np.pi ** 2 * Rij ** 3))


def _dip_tensor(d, r, rotation_axis=None):
    """Full dipolar tensor given a constant and a connecting vector"""

    r = np.array(r)
    r /= np.linalg.norm(r)

    if rotation_axis is None:
        D = d * (3 * r[:, None] * r[None, :] - np.eye(3))
    else:
        a = np.array(rotation_axis)
        a /= np.linalg.norm(a)
        vp2 = np.dot(r, a) ** 2
        D = 0.5 * d * (3 * vp2 - 1) * (3 * a[:, None] * a[None, :] - np.eye(3))

    return D


def _J_constant(Kij, gi, gj):
    """J coupling constants for pairs ij, with reduced constant Kij and
    gyromagnetic ratios gi and gj"""

    return cnst.h * gi * gj * Kij / (4 * np.pi ** 2) * 1e19

def _matrix_to_euler(R:Union[List[List[float]], np.ndarray],
                     convention: str = "zyz",
                     passive: bool = False
                     ) -> np.ndarray:
    """Convert a rotation matrix to Euler angles (in radians)
    
    We use the scipy Rotation class to do this, but we need to make sure that
    the angles are in the conventional ranges for NMR. This function does that.

    Note that SciPy uses the convention that upper case letters are extrinsic rotations
    and lower case letters are intrinsic rotations (note this distinction
    is not that same as active vs passive. We enforce extrinsic rotations here
    by converting all to upper case.

    We use an explicit keyword argument to specify passive rotations, and
    transpose the rotation matrix if necessary.

    Args:
        R (Union[List[List[float]], np.ndarray]): Rotation matrix. Note that SciPy
                                                    will convert this to a proper rotation matrix
                                                    if it is not already one.
                                                    (i.e. det(R) = 1)
        convention (str, optional): Euler angle convention. Defaults to "zyz".
                                    This will be converted to upper case to enforce
                                    extrinsic rotations.
        passive (bool, optional): Whether the angles are passive rotations. Defaults to False.

    Returns:
        np.ndarray: Euler angles in radians

    """
    convention = convention.upper()
    R = np.array(R)
    # (Note that SciPy handles converting to proper rotation matrices)
    Rot = Rotation.from_matrix(R)
    R = Rot.as_matrix() # just in case it was converted

    # If passive, we need to transpose the matrix
    if passive:
        Rot = Rot.inv()
    # Use scipy to get the euler angles
    euler_angles = Rot.as_euler(convention, degrees=False)
    # Now we need to make sure the angles are in the right range
    # for NMR
    euler_angles = _normalise_euler_angles(euler_angles, passive=passive)

    return euler_angles

def _test_euler_rotation(
        euler_angles: np.ndarray,
        eigenvalues: np.ndarray,
        eigenvecs: np.ndarray,
        convention: str = "zyz",
        passive: bool = False,
        eps: float=1e-6)->bool:
    """
    Test that the Euler angles correctly rotate the tensor.

    We compare the tensor rotated by the Euler angles to that you get by
    rotating the tensor with the rotation matrix corresponding to the Euler
    angles. i.e. 

    PAS = np.diag(eigenvalues)
    R = Rotation.from_euler(convention, euler_angles).as_matrix()
    A_rot = np.dot(R, np.dot(PAS, R.T))

    B_rot = np.dot(eigenvecs, np.dot(PAS, eigenvecs.T))

    Args:
        euler_angles (np.ndarray): Euler angles in radians
        eigenvalues (np.ndarray): Eigenvalues of the tensor
        eigenvecs (np.ndarray): Eigenvectors of the tensor
        convention (str, optional): Euler angle convention. Defaults to "zyz".
        passive (bool, optional): Whether the angles are passive rotations. Defaults to False.
        eps (float, optional): Tolerance for degeneracy. Defaults to 1e-6.

    Returns:
        bool: True if the Euler angles correctly rotate the tensor. False otherwise.
    """

    PAS = np.diag(eigenvalues)
    Rot = Rotation.from_euler(convention.upper(), euler_angles)
    if passive:
        Rot = Rot.inv()
    R = Rot.as_matrix()
    A_rot = np.dot(R, np.dot(PAS, np.linalg.inv(R)))
    # B_rot should just be the symmetric tensor
    B_rot = np.linalg.multi_dot([eigenvecs, PAS, np.linalg.inv(eigenvecs)])
    return np.allclose(A_rot, B_rot, atol=eps)





def _handle_euler_edge_cases(
        euler_angles: np.ndarray,
        eigenvalues: np.ndarray,
        original_tensor: np.ndarray,
        convention: str = "zyz",
        passive: bool = False,
        eps: float=1e-6):
    """Handle edge cases in the Euler angle conventions for degenerate tensors.

    Args:
        euler_angles (np.ndarray): Euler angles in radians
        eigenvalues (np.ndarray): Eigenvalues of the tensor
        original_tensor (np.ndarray): Original symmetric tensor
        convention (str, optional): Euler angle convention. Defaults to "zyz".
        eps (float, optional): Tolerance for degeneracy. Defaults to 1e-6.
    """
    A = original_tensor.copy() # alias for convenience

    # only handle zyz or zxz for now
    if convention.lower() not in ["zyz", "zxz"]:
        warnings.warn(
            f"Edge cases not handled for {convention} convention. "
            "Returning unmodified Euler angles."
        )
        return euler_angles

    # Check for degeneracy
    degeneracy = np.sum(np.abs(eigenvalues - eigenvalues[0]) < eps)
    e1, e2, e3 = eigenvalues
    if degeneracy == 1:
        # No degeneracy, just check that we're in the right range
        euler_angles = _normalise_euler_angles(euler_angles, passive=passive)
        return euler_angles

    # this is the tricky one - doubly degenerate
    elif degeneracy == 2:
        if np.abs(e1 - e2) < eps:
            # We have the unique axis along z
            # we are free to set gamma to zero
            euler_angles[2] = 0

        elif np.abs(e2 - e3) < eps:
            # We have the unique axis along x
            # we are free to set alpha to zero
            euler_angles[0] = 0

            # But now we have to be careful
            if convention.lower() == "zyz":
                gamma = np.arcsin(np.sqrt((A[1,1] - e2) / (e1 - e2) )) # +/- this
                gamma = abs(gamma) # we can choose the sign to be positive
                if abs(gamma - np.pi/2) < eps:
                    # We're free to choose beta to be zero
                    beta = 0.0
                elif (np.abs([A[1,2], A[0,1]]) < eps).all():
                    beta = np.arcsin(np.sqrt( (A[2,2] - e3) / (e1 - e3 + (e2 - e1)*np.sin(gamma)**2) ))
                    beta = abs(beta) # we can choose the sign to be positive
                else:
                    beta = np.arctan2(
                        -1*A[1,2] / (np.sin(gamma) * np.cos(gamma)*(e1 - e2)),
                           A[0,1] / (np.sin(gamma) * np.cos(gamma)*(e1 - e2))
                        )
                # Done with zyz
                euler_angles[1] = beta
                euler_angles[2] = gamma

                if passive:
                    a,b,c = euler_angles
                    euler_angles = _normalise_euler_angles(
                                        np.array([-c,-b,-a]),
                                        passive=passive)
                return euler_angles

            if convention.lower() == "zxz":
                gamma = np.arcsin(np.sqrt((A[0,0] - e1) / (e2 - e1) )) # +/- this
                gamma = abs(gamma) # we can choose the sign to be positive
                if abs(gamma) < eps or abs(gamma - np.pi) < eps:
                    # We're free to choose beta to be zero
                    beta = 0.0
                elif (np.abs([A[2,0], A[1,0]]) < eps).all():
                    beta = np.arcsin(
                                np.sqrt(
                                    (A[1,1] - e2 + (e2 - e1)*np.sin(gamma)**2) /
                                    (e3 - e2 + (e2 - e1)*np.sin(gamma)**2)
                                )
                            )
                    beta = abs(beta) # we can choose the sign to be positive
                else:
                    beta = np.arctan2(
                        A[2,0] / (np.sin(gamma) * np.cos(gamma)*(e1 - e2)),
                        A[1,0] / (np.sin(gamma) * np.cos(gamma)*(e1 - e2))
                        )
                # Done with zxz
                euler_angles[1] = beta
                euler_angles[2] = gamma

                if passive:
                    a,b,c = euler_angles
                    euler_angles = _normalise_euler_angles(
                                        np.array([-c,-b,-a]),
                                        passive=passive)
                return euler_angles
        else:
            # We shouldn't have the unique axis along y for
            # reasonably sorted eigenvalues
            raise ValueError("Unexpected degeneracy when computing Euler angles.\n"
                             "Eigenvalues are ordered: ", eigenvalues)
    elif degeneracy == 3:
        # All eigenvalues are the same
        # We can set all angles to zero
        euler_angles = np.zeros(3)
    else:
        raise ValueError("Degeneracy must be 1, 2, or 3.")
    return euler_angles


def _normalise_euler_angles(
        euler_angles: np.ndarray,
        passive: bool = False,
        eps: float = 1e-6,
        )-> np.ndarray:
    """
    Normalise Euler angles to standard ranges for NMR as 
    defined in: 
    TensorView for MATLAB :cite:p:`Svenningsson2023`

    Args:
        euler_angles (np.ndarray): Euler angles in radians
        passive (bool, optional): Whether the angles are passive rotations. Defaults to False.
        eps (float, optional): Tolerance for degeneracy. Defaults to 1e-6.



    .. footbibliography::
    
    """
    alpha, beta, gamma = euler_angles

    # wrap any negative angles
    alpha = alpha % (2 * np.pi)
    beta  = beta  % (2 * np.pi)
    gamma = gamma % (2 * np.pi)


    if passive:
        # Note this assumes we've already transposed the R matrix
        if beta > np.pi:
            beta = 2 * np.pi - beta
            gamma = gamma - np.pi
            gamma  = gamma % (2 * np.pi)

        if beta >= np.pi/2 - eps:
            alpha = np.pi - alpha
            alpha  = alpha % (2 * np.pi)
            beta = np.pi - beta
            beta = beta % (2 * np.pi)
            gamma = np.pi + gamma
            gamma = gamma % (2 * np.pi)

        if alpha >= np.pi - eps:
            alpha = alpha - np.pi
    else:
        if beta > np.pi:
            beta = 2 * np.pi - beta
            alpha = alpha - np.pi
            alpha  = alpha % (2 * np.pi)

        if beta >= np.pi/2 - eps:
            alpha = alpha + np.pi
            alpha  = alpha % (2 * np.pi)
            beta = np.pi - beta
            beta = beta % (2 * np.pi)
            gamma = np.pi - gamma
            gamma = gamma % (2 * np.pi)

        if gamma >= np.pi - eps:
            gamma = gamma - np.pi

    return np.array([alpha, beta, gamma])


def _equivalent_euler(euler_angles: np.ndarray, passive: bool = False):
    """
    Find the equivalent Euler angles for a given set of Euler angles.

    This set should be correct for NMR tensors, according to 
    TensorView for MATLAB: :cite:p:`Svenningsson2023`
    """

    equiv_angles = np.zeros((4, 3))

    alpha, beta, gamma = euler_angles

    # set the first row of the array to the original Euler angles
    equiv_angles[0] = [alpha, beta, gamma]
    # the order of these doesn't matter, but has been chosen to match
    # that in the TensorView for MATLAB code
    # (which is different to that in the corresponding paper)
    if passive:
        equiv_angles[1,:] = [np.pi + alpha, beta, gamma]
        equiv_angles[2,:] = [np.pi - alpha, np.pi - beta, np.pi + gamma]
        equiv_angles[3,:] = [2*np.pi - alpha, np.pi - beta,np.pi + gamma]
    else:
        equiv_angles[1,:] = [alpha, beta, np.pi + gamma]
        equiv_angles[2,:] = [np.pi + alpha, np.pi - beta, np.pi - gamma]
        equiv_angles[3,:] = [np.pi + alpha, np.pi - beta, 2*np.pi - gamma]

    # now a few checks
    # wrap any negative angles
    equiv_angles[equiv_angles < 0] = equiv_angles[equiv_angles < 0] % (2*np.pi)
    # wrap any values > 2pi
    equiv_angles[equiv_angles >= 2*np.pi] = equiv_angles[equiv_angles >= 2*np.pi] % (2*np.pi)

    return equiv_angles

def _equivalent_relative_euler(euler_angles: np.ndarray, passive: bool = False) -> np.ndarray:
    """
    Returns a list of 16 equivalent relative Euler angles for a given set of Euler angles that corresponds to the relative orientation of two NMR tensors.
    See TensorView for MATLAB: :cite:p:`Svenningsson2023`
    """
    equiv_angles = np.zeros((16, 3))
    alpha, beta, gamma = euler_angles

    if passive:
        # --- Passive ZYZ or ZXZ --- #
        equiv_angles[0,:] = [alpha, beta, gamma]
        equiv_angles[1,:] = [2*np.pi - alpha, np.pi - beta, np.pi + gamma]
        equiv_angles[2,:] = [np.pi - alpha, np.pi - beta, np.pi + gamma]
        equiv_angles[3,:] = [np.pi + alpha, beta, gamma]
        equiv_angles[4,:] = [np.pi + alpha, np.pi - beta, 2*np.pi - gamma]
        equiv_angles[5,:] = [np.pi - alpha, beta, np.pi - gamma]
        equiv_angles[6,:] = [2*np.pi - alpha, beta, np.pi - gamma]
        equiv_angles[7,:] = [alpha, np.pi - beta, 2*np.pi - gamma]
        equiv_angles[8,:] = [np.pi + alpha, np.pi - beta, np.pi - gamma]
        equiv_angles[9,:] = [np.pi - alpha, beta, 2*np.pi - gamma]
        equiv_angles[10,:] = [2*np.pi - alpha, beta,2*np.pi - gamma]
        equiv_angles[11,:] = [alpha, np.pi - beta, np.pi - gamma]
        equiv_angles[12,:] = [alpha, beta, np.pi + gamma]
        equiv_angles[13,:] = [2*np.pi - alpha, np.pi - beta, gamma]
        equiv_angles[14,:] = [np.pi - alpha, np.pi - beta, gamma]
        equiv_angles[15,:] = [np.pi + alpha, beta, np.pi + gamma]
    else:
        # --- Active ZYZ or ZXZ--- #
        equiv_angles[0,:] = [alpha, beta, gamma]
        equiv_angles[1,:] = [alpha + np.pi, np.pi - beta, 2*np.pi - gamma]
        equiv_angles[2,:] = [alpha + np.pi, np.pi - beta, np.pi - gamma]
        equiv_angles[3,:] = [alpha, beta, np.pi + gamma]
        equiv_angles[4,:] = [2*np.pi - alpha, np.pi - beta, np.pi + gamma]
        equiv_angles[5,:] = [np.pi - alpha, beta, np.pi - gamma]
        equiv_angles[6,:] = [np.pi - alpha, beta, 2*np.pi - gamma]
        equiv_angles[7,:] = [2*np.pi - alpha, np.pi - beta, gamma]
        equiv_angles[8,:] = [np.pi - alpha, np.pi - beta, np.pi + gamma]
        equiv_angles[9,:] = [2*np.pi - alpha, beta, np.pi - gamma]
        equiv_angles[10,:] = [2*np.pi - alpha, beta,2*np.pi - gamma]
        equiv_angles[11,:] = [np.pi - alpha, np.pi - beta, gamma]
        equiv_angles[12,:] = [alpha + np.pi, beta, gamma]
        equiv_angles[13,:] = [alpha, np.pi - beta, 2*np.pi - gamma]
        equiv_angles[14,:] = [alpha, np.pi - beta, np.pi - gamma]
        equiv_angles[15,:] = [alpha + np.pi, beta, np.pi + gamma]



    # wrap any negative angles
    equiv_angles[equiv_angles < 0] = equiv_angles[equiv_angles < 0] % (2*np.pi)
    # wrap any values >= 2pi
    equiv_angles[equiv_angles >= 2*np.pi] = equiv_angles[equiv_angles >= 2*np.pi] % (2*np.pi)


    return equiv_angles


def _tryallanglestest(
        euler_angles: np.ndarray,
        pas1: np.ndarray,
        pasv2: np.ndarray,
        arel1: np.ndarray,
        convention: str,
        eps: float = 1e-3
        ) -> np.ndarray:
    """
    For relative Euler angles from tensor A to B, if B is axially symmetric,
    we need to try all equivalent Euler angles of the symmetric tensor to
    find the conventional one. 

    Go through the 4 equivalent passive angles since we're using 
    the trick of calculating the A relative angles in the frame of B 
    by calculating the B relative angles in the frame of A with the *passive* convention.

    Credit: function adapted from TensorView for MATLAB: :cite:p:`Svenningsson2023`
    """

    # make copy of the input angles
    euler_angles_out = euler_angles.copy()
    rrel_check = Rotation.from_euler(convention.upper(), euler_angles).as_matrix()
    mcheck = np.round(np.dot(np.dot(rrel_check, pas1), np.linalg.inv(rrel_check)), 14)
    # Define the ways in which the angles should be updated
    alpha, beta, gamma = euler_angles
    updates = [
        lambda: (alpha + np.pi, beta, gamma),
        lambda: (2*np.pi - alpha, np.pi - beta, gamma + np.pi),
        lambda: (np.pi-alpha, np.pi-beta, gamma + np.pi),
    ]
    # if pasv2[0] == pasv2[1]:
    if np.allclose(pasv2[0], pasv2[1], atol=eps):
        # Iterate over the updates, updating only if the angles don't match
        for update in updates:
            if not np.all(np.isclose(arel1, mcheck, atol=eps)):
                alpha_out, beta_out, gamma_out = update()
                euler_angles_out, mcheck = _compute_rotation([alpha_out, beta_out, gamma_out], arel1, pas1, convention, 1)
            else:
                break # If the angles match, we're done

        # If the last condition is still true, print a message
        if not np.all(np.isclose(arel1, mcheck, atol=eps)):
            raise 'Failed isequal check at (_tryallanglestest) please contact the developers for to help resolve this issue.'

    elif np.allclose(pasv2[0], pasv2[1], atol=eps):
        # Iterate over the updates, updating only if the angles don't match
        for update in updates:
            if not np.all(np.isclose(arel1, mcheck, atol=eps)):
                alpha_out, beta_out, gamma_out = update()
                euler_angles_out, mcheck = _compute_rotation([alpha_out, beta_out, gamma_out], arel1, pas1, convention, 2)
            else:
                break

    return euler_angles_out



def _compute_rotation(euler_angles: np.ndarray,
                      arel1: np.ndarray,
                      pas1: np.ndarray,
                      convention: str,
                      rotation_type: int
                      )-> Tuple[np.ndarray, np.ndarray]:
    rrel_check = Rotation.from_euler(convention.upper(), euler_angles).as_matrix()
    mcheck = np.round(np.dot(np.dot(rrel_check, pas1), np.linalg.inv(rrel_check)), 14)

    if rotation_type == 1:
        component1 = arel1[2, 0] + arel1[2, 1]*mcheck[2, 1]/mcheck[2, 0]
        component2 = mcheck[2, 0] + mcheck[2, 1]*mcheck[2, 1]/mcheck[2, 0]
        component3 = arel1[2, 1] - arel1[2, 0]*mcheck[2, 1]/mcheck[2, 0]
        component4 = mcheck[2, 0] + mcheck[2, 1]*mcheck[2, 1]/mcheck[2, 0]
        euler_convention = "ZYZ"
    elif rotation_type == 2:
        component1 = arel1[2, 0] + arel1[1, 0]*mcheck[1, 0]/mcheck[2, 0]
        component2 = mcheck[2, 0] + mcheck[1, 0]*mcheck[1, 0]/mcheck[2, 0]
        component3 = arel1[2, 0] - arel1[1, 0]*mcheck[2, 0]/mcheck[1, 0]
        component4 = mcheck[1, 0] + mcheck[2, 0]*mcheck[2, 0]/mcheck[1, 0]
        euler_convention = "ZXZ"

    symrotang_check = np.arctan2(component3/component4, component1/component2)
    symrot_check = Rotation.from_euler(euler_convention, [0, 0, symrotang_check]).as_matrix()
    mcheck = np.dot(np.dot(symrot_check, mcheck), np.linalg.inv(symrot_check))
    return euler_angles, mcheck
def _frange(a, b, x):
    while a < b:
        yield a
        a += x

def _ensure_tensor_format(tensors):
    """
    Ensure tensors are in 3x3 format, converting from flattened 1x9 if necessary.
    
    Args:
        tensors (np.ndarray): Tensor array, either (N, 3, 3) or (N, 9)
        
    Returns:
        np.ndarray: Tensors in (N, 3, 3) format
    """
    tensors = np.array(tensors)
    
    if tensors.ndim == 2 and tensors.shape[1] == 9:
        return tensors.reshape(-1, 3, 3)
    elif tensors.ndim == 3 and tensors.shape[1:] == (3, 3):
        return tensors
    else:
        raise ValueError(f"Invalid tensor format: expectd (N, 9) or (N, 3, 3), got {tensors.shape}")

def _get_tensor_array(s, tensor_tag):
    """
    Get tensor array from structure, ensuring 3x3 format.
    
    Args:
        s (Atoms): The structure
        tensor_tag (str): Name of the tensor array
        
    Returns:
        np.ndarray: Tensor array in (N, 3, 3) format
        
    Raises:
        RuntimeError: If tensor array not found
    """
    if not s.has(tensor_tag):
        raise RuntimeError(f"Structure does not have tensor array '{tensor_tag}'")
    
    tensors = s.get_array(tensor_tag)
    return _ensure_tensor_format(tensors)
